import torch

from PIL import Image
from tqdm import tqdm
from torch import Tensor
from typing import List
from typing import Tuple
from cfcv.misc.toolkit import to_rgb

from .utils import predict_folder
from ..data import SSLTestTransform
from ..data import InferenceImageFolderData
from ..pipeline import SimplePipeline
from ....constants import LATENT_KEY
from ....misc.toolkit import to_torch
from ....misc.toolkit import to_device
from ....misc.toolkit import eval_context


class DINOExtractor:
    def __init__(self, m: SimplePipeline, img_size: int, *, to_gray: bool = False):
        self.m = m
        self.dino = m.model
        self.transform = SSLTestTransform(img_size, to_gray)

    @property
    def device(self) -> torch.device:
        return self.dino.device

    def get_latent(self, src_path: str) -> Tensor:
        src = to_rgb(Image.open(src_path))
        net = to_torch(self.transform(src)[None, ...]).to(self.device)
        with eval_context(self.dino):
            return self.dino.get_latent(net)

    def get_logits(self, src_path: str) -> Tensor:
        src = to_rgb(Image.open(src_path))
        net = self.transform(src)[None, ...].to(self.device)
        with eval_context(self.dino):
            return self.dino.get_logits(net)

    def get_folder_latent(
        self,
        src_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        use_tqdm: bool = True,
    ) -> Tuple[Tensor, List[str]]:
        results = predict_folder(
            self.m,
            src_folder,
            batch_size=batch_size,
            num_workers=num_workers,
            transform=self.transform,
            use_tqdm=use_tqdm,
        )
        latent = to_torch(results.outputs[LATENT_KEY])
        return latent, results.img_paths

    def get_folder_logits(
        self,
        src_folder: str,
        *,
        batch_size: int,
        num_workers: int = 0,
        use_tqdm: bool = True,
    ) -> Tuple[Tensor, List[str]]:
        data = InferenceImageFolderData(
            src_folder,
            batch_size=batch_size,
            num_workers=num_workers,
            transform=self.transform,
        )
        outputs = []
        iterator = data.initialize()[0]
        if use_tqdm:
            iterator = tqdm(iterator, total=len(iterator))
        with eval_context(self.dino):
            for i, batch in enumerate(iterator):
                batch = to_device(batch, self.device)
                outputs.append(self.dino.student(i, batch).cpu())
        return torch.cat(outputs, dim=0), data.dataset.img_paths


__all__ = [
    "DINOExtractor",
]



import torch

import torch.nn as nn

from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Optional
from typing import NamedTuple
from cftool.misc import WithRegister

from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Conv2d


discriminator_dict: Dict[str, Type["DiscriminatorBase"]] = {}


class DiscriminatorOutput(NamedTuple):
    output: torch.Tensor
    cond_logits: Optional[torch.Tensor] = None


class DiscriminatorBase(nn.Module, WithRegister["DiscriminatorBase"]):
    d = discriminator_dict

    clf: nn.Module
    net: nn.Module
    cond: Optional[nn.Module]

    def __init__(
        self,
        in_channels: int,
        num_classes: Optional[int] = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.num_classes = num_classes

    def generate_cond(self, out_channels: int) -> None:
        if self.num_classes is None:
            self.cond = None
        else:
            self.cond = Conv2d(
                out_channels,
                self.num_classes,
                kernel_size=4,
                padding=1,
                stride=1,
            )

    def forward(self, net: torch.Tensor) -> Any:
        feature_map = self.net(net)
        logits = self.clf(feature_map)
        cond_logits = None
        if self.cond is not None:
            cond_logits_map = self.cond(feature_map)
            cond_logits = torch.mean(cond_logits_map, [2, 3])
        return DiscriminatorOutput(logits, cond_logits)


@DiscriminatorBase.register("basic")
class NLayerDiscriminator(DiscriminatorBase):
    def __init__(
        self,
        in_channels: int,
        num_classes: Optional[int] = None,
        *,
        num_layers: int = 2,
        start_channels: int = 16,
        norm_type: Optional[str] = "batch",
    ):
        super().__init__(in_channels, num_classes)
        self.num_layers = num_layers
        self.start_channels = start_channels
        # backbone
        blocks: List[nn.Module] = [
            Conv2d(
                in_channels,
                self.start_channels,
                kernel_size=4,
                stride=2,
                padding=1,
                bias=False,
            ),
            nn.LeakyReLU(0.2, inplace=True),
        ]
        nc_multiplier = 1
        for i in range(1, num_layers):
            nc_multiplier_prev = nc_multiplier
            nc_multiplier = min(2**i, 8)
            blocks.extend(
                get_conv_blocks(
                    start_channels * nc_multiplier_prev,
                    start_channels * nc_multiplier,
                    4,
                    2,
                    bias=False,
                    padding=1,
                    norm_type=norm_type,
                    activation=nn.LeakyReLU(0.2, inplace=True),
                )
            )
        self.net = nn.Sequential(*blocks)
        # heads
        out_channels = start_channels * nc_multiplier
        self.clf = Conv2d(
            out_channels,
            1,
            kernel_size=4,
            padding=1,
            stride=1,
            bias=False,
        )
        self.generate_cond(out_channels)


@DiscriminatorBase.register("multi_scale")
class MultiScaleDiscriminator(DiscriminatorBase):
    def __init__(
        self,
        in_channels: int,
        num_classes: Optional[int] = None,
        *,
        num_scales: int = 3,
        num_layers: int = 4,
        latent_channels: int = 64,
        activation: str = "leaky_relu_0.2",
        norm_type: Optional[str] = None,
    ):
        if num_classes is not None:
            msg = "`MultiScaleDiscriminator` does not support conditional inputs"
            raise ValueError(msg)
        super().__init__(in_channels, num_classes)
        self.num_layers = num_layers
        self.latent_channels = latent_channels
        self.activation = activation
        self.norm_type = norm_type
        self.downsample = nn.AvgPool2d(
            3,
            stride=2,
            padding=[1, 1],
            count_include_pad=False,
        )
        nets = [self._make() for _ in range(num_scales)]
        self.discriminators = nn.ModuleList(nets)

    def _make(self) -> nn.Sequential:
        in_nc = self.in_channels
        out_nc = self.latent_channels
        blocks = []
        for i in range(self.num_layers):
            blocks.extend(
                get_conv_blocks(
                    in_nc,
                    out_nc,
                    4,
                    2,
                    norm_type=None if i == 0 else self.norm_type,
                    activation=self.activation,
                    padding="reflection1",
                )
            )
            in_nc = out_nc
            out_nc *= 2
        blocks.append(Conv2d(in_nc, 1, kernel_size=1, stride=1, padding=0))
        return nn.Sequential(*blocks)

    def forward(self, net: torch.Tensor) -> List[DiscriminatorOutput]:
        outputs = []
        for discriminator in self.discriminators:
            outputs.append(DiscriminatorOutput(discriminator(net)))
            net = self.downsample(net)
        return outputs


__all__ = [
    "DiscriminatorOutput",
    "DiscriminatorBase",
    "NLayerDiscriminator",
    "MultiScaleDiscriminator",
]



import torch

import torch.nn as nn

from abc import abstractmethod
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Callable
from typing import Optional
from torch.cuda.amp import GradScaler
from torch.optim.optimizer import Optimizer

from ...types import losses_type
from ...types import tensor_dict_type
from ...protocol import StepOutputs
from ...protocol import LossProtocol
from ...protocol import TrainerState
from ...protocol import ModelProtocol
from ...protocol import MetricsOutputs
from ...protocol import DataLoaderProtocol
from ...protocol import ModelWithCustomSteps
from ...constants import LOSS_KEY
from ...constants import INPUT_KEY
from ...constants import PREDICTIONS_KEY


def register_module(
    name: str,
    *,
    pre_bases: Optional[List[type]] = None,
    post_bases: Optional[List[type]] = None,
) -> Callable[[Type[nn.Module]], Type[nn.Module]]:
    def _core(m: Type[nn.Module]) -> Type[nn.Module]:
        @ModelProtocol.register(name)
        class _(*bases):  # type: ignore
            def __init__(self, *args: Any, **kwargs: Any):
                super().__init__()
                self.core = m(*args, **kwargs)

            def forward(
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                state: Optional[TrainerState] = None,
                **kwargs: Any,
            ) -> tensor_dict_type:
                rs = self.core(batch[INPUT_KEY], **kwargs)
                if not isinstance(rs, dict):
                    rs = {PREDICTIONS_KEY: rs}
                return rs

        return m

    bases = (pre_bases or []) + [ModelProtocol] + (post_bases or [])
    return _core


def _get_clip_norm_fn(trainer: Any) -> Callable[[], None]:
    def _core() -> None:
        if trainer.clip_norm > 0.0:
            trainer.clip_norm_step()

    return _core


class CustomModule(nn.Module, metaclass=ABCMeta):
    @property
    def device(self) -> torch.device:
        return list(self.parameters())[0].device

    @abstractmethod
    def train_step(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        optimizers: Dict[str, Optimizer],
        use_amp: bool,
        grad_scaler: GradScaler,
        clip_norm_fn: Callable[[], None],
        scheduler_step_fn: Callable[[], None],
        trainer: Any,
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> StepOutputs:
        pass

    @abstractmethod
    def evaluate_step(
        self,
        loader: DataLoaderProtocol,
        portion: float,
        weighted_loss_score_fn: Callable[[Dict[str, float]], float],
        trainer: Any,
    ) -> MetricsOutputs:
        pass

    @abstractmethod
    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        pass

    @staticmethod
    def params_groups(m: nn.Module) -> Any:
        pass

    def init_ddp(self) -> None:
        pass

    def permute_trainer_config(self, trainer_config: Dict[str, Any]) -> None:
        pass


def register_custom_module(
    name: str,
    *,
    pre_bases: Optional[List[type]] = None,
    post_bases: Optional[List[type]] = None,
    custom_train_step: bool = True,
    custom_evaluate_step: bool = True,
    custom_params_groups: bool = False,
    custom_ddp_initialization: bool = False,
) -> Callable[[Type[CustomModule]], Type[CustomModule]]:
    def _core(m: Type[CustomModule]) -> Type[CustomModule]:
        @ModelProtocol.register(name)
        class _(*bases):  # type: ignore
            def __init__(self, *args: Any, **kwargs: Any):
                super().__init__()
                self.core = m(*args, **kwargs)

            def forward(
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                state: Optional[TrainerState] = None,
                **kwargs: Any,
            ) -> tensor_dict_type:
                return self.core.forward(batch_idx, batch, state, **kwargs)

            def train_step(
                self,
                batch_idx: int,
                batch: tensor_dict_type,
                trainer: Any,
                forward_kwargs: Dict[str, Any],
                loss_kwargs: Dict[str, Any],
            ) -> StepOutputs:
                return self.core.train_step(
                    batch_idx,
                    batch,
                    trainer.optimizers,
                    trainer.use_amp,
                    trainer.grad_scaler,
                    _get_clip_norm_fn(trainer),
                    trainer.scheduler_step,
                    trainer,
                    forward_kwargs,
                    loss_kwargs,
                )

            def evaluate_step(  # type: ignore
                self,
                loader: DataLoaderProtocol,
                portion: float,
                trainer: Any,
            ) -> MetricsOutputs:
                return self.core.evaluate_step(
                    loader,
                    portion,
                    trainer.weighted_loss_score,
                    trainer,
                )

            def params_groups(self, m_: nn.Module) -> Any:
                return self.core.params_groups(m_)

            def init_ddp(self) -> None:
                self.core.init_ddp()

            def permute_trainer_config(self, trainer_config: Dict[str, Any]) -> None:
                self.core.permute_trainer_config(trainer_config)

        _.custom_train_step = custom_train_step
        _.custom_evaluate_step = custom_evaluate_step
        _.custom_params_groups = custom_params_groups
        _.custom_ddp_initialization = custom_ddp_initialization
        return m

    bases = (pre_bases or []) + [ModelWithCustomSteps] + (post_bases or [])
    return _core


class LossModule(nn.Module, metaclass=ABCMeta):
    @abstractmethod
    def forward(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        pass


def register_loss_module(name: str) -> Callable[[Type[LossModule]], Type[LossModule]]:
    def _core(loss_base: Type[LossModule]) -> Type[LossModule]:
        @LossProtocol.register(name)
        class _(LossProtocol):  # type: ignore
            def __init__(self, reduction: str = "mean", **kwargs: Any):
                super().__init__(reduction, **kwargs)
                self.core = loss_base(**kwargs)

            def _core(
                self,
                forward_results: tensor_dict_type,
                batch: tensor_dict_type,
                state: Optional[TrainerState] = None,
                **kwargs: Any,
            ) -> losses_type:
                rs = self.core(forward_results, batch, state, **kwargs)
                if isinstance(rs, Tensor):
                    rs = {LOSS_KEY: rs}
                return rs

        return loss_base

    return _core


__all__ = [
    "register_module",
    "register_custom_module",
    "register_loss_module",
    "CustomModule",
    "LossModule",
]



import numpy as np

from typing import Dict
from typing import Callable

from cftool.array import contrast_noise
from cftool.array import fractal_noise_2d
from cfcv.misc.toolkit import min_max_normalize


noise_fn_type = Callable[[int, int], np.ndarray]
noises: Dict[str, noise_fn_type] = {}


def register_noise(name: str) -> Callable[[noise_fn_type], noise_fn_type]:
    def _core(fn: noise_fn_type) -> noise_fn_type:
        noises[name] = fn
        return fn

    return _core


@register_noise("fractal")
def fractal_noise(w: int, h: int) -> np.ndarray:
    if w > 1024 or h > 1024:
        side, octaves = 2048, 7
    elif w > 512 or h > 512:
        side, octaves = 1024, 6
    elif w > 256 or h > 256:
        side, octaves = 512, 5
    else:
        side, octaves = 256, 4

    r = min_max_normalize(fractal_noise_2d((side, side), (32, 32), octaves))
    g = min_max_normalize(fractal_noise_2d((side, side), (32, 32), octaves))
    b = min_max_normalize(fractal_noise_2d((side, side), (32, 32), octaves))
    stack = np.dstack((contrast_noise(r), contrast_noise(g), contrast_noise(b)))
    return (255.0 * stack).astype(np.uint8)


__all__ = ["noises"]



import os
import json
import shutil

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Callable
from typing import Optional

from ..types import tensor_dict_type
from ..trainer import DeviceInfo
from ..api.cv.pipeline import SimplePipeline
from ..api.cv.models.interface import IImageExtractor

try:
    import faiss
except ImportError:
    from ..constants import WARNING_PREFIX

    print(f"{WARNING_PREFIX}`faiss` need to be installed to use faiss based scripts")


def run_faiss(
    x: np.ndarray,
    index_path: str,
    *,
    dimension: int,
    factory: str = "IVF128,Flat",
) -> Any:
    index = faiss.index_factory(dimension, factory)
    print(">> training index")
    index.train(x)
    print(">> adding data to index")
    index.add(x)
    print(">> saving index")
    faiss.write_index(index, index_path)
    print("> done")


def image_retrieval(
    m: SimplePipeline,
    packed: str,
    extractor: IImageExtractor,
    *,
    tag: str,
    task: str,
    data_folder: str,
    input_sample: tensor_dict_type,
    index_dimension: int,
    path_converter: Callable[[str], str],
    batch_size: int = 128,
    num_workers: int = 32,
    index_factory: str = "IVF128,Flat",
    forward_fn: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
    output_names: Optional[List[str]] = None,
    cuda: int = 7,
) -> None:
    version_folder = os.path.join(".versions", task, tag)
    features_folder = os.path.join(version_folder, "features")
    dist_folder = os.path.join(version_folder, "dist")
    model_file = f"{tag}.zip"
    onnx_file = f"{tag}.onnx"
    features_file = "features.npy"
    files_file = "files.json"

    if os.path.isdir(version_folder):
        print(f"> Warning : '{version_folder}' already exists, it will be removed")
        shutil.rmtree(version_folder)
    m.device_info = DeviceInfo(str(cuda), None)
    m.model.to(f"cuda:{cuda}")
    m.to_onnx(
        version_folder,
        onnx_file=onnx_file,
        forward_fn=forward_fn,
        input_sample=input_sample,
        output_names=output_names,
    )

    os.makedirs(features_folder, exist_ok=True)
    kw = dict(batch_size=batch_size, num_workers=num_workers)
    files = []
    features = []
    for split in ["train", "valid"]:
        split_folder = os.path.join(data_folder, split)
        with open(os.path.join(split_folder, "path_mapping.json"), "r") as f:
            mapping = json.load(f)
        rs = extractor.get_folder_latent(split_folder, **kw)  # type: ignore
        files.extend([mapping[file] for file in rs[1]])
        features.append(rs[0])

    x = np.vstack(features)
    np.save(os.path.join(features_folder, features_file), x)
    files_path = os.path.join(features_folder, files_file)
    with open(files_path, "w") as f:
        json.dump([path_converter(file) for file in files], f, ensure_ascii=False)

    index = faiss.index_factory(index_dimension, index_factory)
    print(">> training index")
    index.train(x)
    print(">> adding data to index")
    index.add(x)
    print(">> saving index")
    index_file = f"{task}.{tag}.index"
    index_path = os.path.join(features_folder, index_file)
    faiss.write_index(index, index_path)
    print("> done")

    os.makedirs(dist_folder)
    print(">> copying model")
    shutil.copyfile(f"{packed}.zip", os.path.join(dist_folder, model_file))
    print(">> copying onnx")
    shutil.copyfile(
        os.path.join(version_folder, onnx_file),
        os.path.join(dist_folder, onnx_file),
    )
    print(">> copying files")
    shutil.copyfile(files_path, os.path.join(dist_folder, files_file))
    print(">> copying index")
    shutil.copyfile(index_path, os.path.join(dist_folder, index_file))
    print("> done")


__all__ = [
    "run_faiss",
    "image_retrieval",
]



import torch.nn as nn

from typing import Any
from typing import Optional

from .protocol import MERGED_KEY
from .protocol import MLCoreProtocol
from ...types import tensor_dict_type
from ...protocol import TrainerState
from ...constants import PREDICTIONS_KEY


@MLCoreProtocol.register("linear")
class Linear(MLCoreProtocol):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_history: int,
        *,
        bias: bool = True,
    ):
        super().__init__(in_dim, out_dim, num_history)
        in_dim *= num_history
        self.net = nn.Linear(in_dim, out_dim, bias)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        net = batch[MERGED_KEY]
        if len(net.shape) > 2:
            net = net.contiguous().view(len(net), -1)
        return {PREDICTIONS_KEY: self.net(net)}


__all__ = ["Linear"]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Optional

from .protocol import DecoderBase
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....constants import INPUT_KEY
from ....constants import PREDICTIONS_KEY
from ..encoder.vqgan import normalize
from ..encoder.vqgan import Swish
from ..encoder.vqgan import AttnBlock
from ..encoder.vqgan import ResidualBlock
from ....misc.toolkit import interpolate
from ....modules.blocks import Conv2d


class Upsample(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.net = Conv2d(dim, dim, kernel_size=3, padding=1)

    def forward(self, net: Tensor) -> Tensor:
        net = interpolate(net, factor=2.0, mode="nearest")
        net = self.net(net)
        return net


@DecoderBase.register("vqgan")
class VQGANDecoder(DecoderBase):
    def __init__(
        self,
        img_size: int,
        out_channels: int = 3,
        latent_channels: int = 256,
        *,
        channel_multipliers: Optional[List[int]] = None,
        attention_resolutions: Optional[List[int]] = None,
        latent_dim: int = 128,
        res_dropout: float = 0.0,
        num_res_blocks: int = 3,
        num_upsample: Optional[int] = None,
        cond_channels: int = 16,
        num_classes: Optional[int] = None,
        latent_resolution: Optional[int] = None,
    ):
        if channel_multipliers is None:
            channel_multipliers = [1, 1, 2, 2, 4]
        if attention_resolutions is None:
            attention_resolutions = [16]
        if num_upsample is not None and num_upsample != len(channel_multipliers):
            raise ValueError(
                f"`num_upsample` ({num_upsample}) should be identical with "
                f"the length of `channel_multipliers` ({len(channel_multipliers)}"
            )
        num_upsample = len(channel_multipliers)
        super().__init__(
            latent_channels,
            out_channels,
            img_size=img_size,
            num_upsample=num_upsample,
            cond_channels=cond_channels,
            num_classes=num_classes,
            latent_resolution=latent_resolution,
        )
        in_nc = latent_dim * channel_multipliers[-1]
        current_res = img_size // 2 ** (num_upsample - 1)
        # in conv
        self.in_conv = Conv2d(latent_channels, in_nc, kernel_size=3, padding=1)
        # mid
        self.mid = nn.Sequential(
            ResidualBlock(in_nc, in_nc, res_dropout),
            AttnBlock(in_nc),
            ResidualBlock(in_nc, in_nc, res_dropout),
        )
        # upsample
        blocks = []
        for i in reversed(range(num_upsample)):
            i_blocks = []
            out_nc = latent_dim * channel_multipliers[i]
            for j in range(num_res_blocks):
                i_blocks.append(ResidualBlock(in_nc, out_nc, res_dropout))
                in_nc = out_nc
                if current_res in attention_resolutions:
                    i_blocks.append(AttnBlock(in_nc))
            if i != 0:
                i_blocks.append(Upsample(in_nc))
                current_res *= 2
            blocks.append(nn.Sequential(*i_blocks))
        self.upsample_blocks = nn.ModuleList(blocks[::-1])
        # finalize
        self.out = nn.Sequential(
            normalize(in_nc),
            Swish(),
            Conv2d(in_nc, out_channels, kernel_size=3, padding=1),
        )

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        net = batch[INPUT_KEY]
        net = self.in_conv(net)
        net = self.mid(net)
        for block in self.upsample_blocks[::-1]:
            net = block(net)
        net = self.out(net)
        if kwargs.get("resize", True):
            net = self.resize(net)
        return {PREDICTIONS_KEY: net}


__all__ = [
    "VQGANDecoder",
]



import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cftool.misc import check_requires

from .vanilla import reparameterize
from ..protocol import GaussianGeneratorMixin
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....protocol import ModelProtocol
from ....constants import LABEL_KEY
from ....constants import PREDICTIONS_KEY
from ..encoder.protocol import Encoder1DBase
from ...implicit.siren import ImgSiren
from ....misc.toolkit import auto_num_layers
from ....losses.vae import MU_KEY
from ....losses.vae import LOG_VAR_KEY
from ....modules.blocks import Linear


@ModelProtocol.register("siren_vae")
class SirenVAE(ModelProtocol, GaussianGeneratorMixin):
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        *,
        min_size: int = 2,
        latent_dim: int = 256,
        target_downsample: Optional[int] = None,
        num_classes: Optional[int] = None,
        conditional_dim: int = 16,
        num_layers: int = 4,
        w_sin: float = 1.0,
        w_sin_initial: float = 30.0,
        bias: bool = True,
        final_activation: Optional[str] = None,
        encoder1d: str = "vanilla",
        encoder1d_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.img_size = img_size
        self.latent_dim = latent_dim
        self.num_classes = num_classes
        self.out_channels = out_channels or in_channels
        args = img_size, min_size, target_downsample
        num_downsample = auto_num_layers(*args, use_stride=encoder1d == "vanilla")
        # encoder
        encoder1d_base = Encoder1DBase.get(encoder1d)
        if encoder1d_config is None:
            encoder1d_config = {}
        if check_requires(encoder1d_base, "img_size"):
            encoder1d_config["img_size"] = img_size
        encoder1d_config["in_channels"] = in_channels
        encoder1d_config["latent_dim"] = latent_dim
        if encoder1d == "vanilla":
            encoder1d_config["num_downsample"] = num_downsample
        self.encoder = encoder1d_base(**encoder1d_config)
        self.to_statistics = Linear(latent_dim, 2 * latent_dim, bias=False)
        # siren
        self.siren = ImgSiren(
            img_size,
            self.out_channels,
            latent_dim,
            num_classes,
            conditional_dim,
            num_layers=num_layers,
            w_sin=w_sin,
            w_sin_initial=w_sin_initial,
            bias=bias,
            final_activation=final_activation,
        )

    @property
    def can_reconstruct(self) -> bool:
        return True

    def decode(
        self,
        z: Tensor,
        *,
        labels: Optional[Tensor],
        size: Optional[int] = None,
        **kwargs: Any,
    ) -> Tensor:
        net = self.siren.decode(z, labels=labels, size=size)
        net = torch.tanh(net)
        return net

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        net = self.encoder.encode(batch, **kwargs)
        net = self.to_statistics(net)
        mu, log_var = net.chunk(2, dim=1)
        net = reparameterize(mu, log_var)
        net = self.siren.decode(net, labels=batch.get(LABEL_KEY))
        return {PREDICTIONS_KEY: net, MU_KEY: mu, LOG_VAR_KEY: log_var}


__all__ = [
    "SirenVAE",
]



import torch.nn as nn

from typing import Any
from typing import List
from typing import Optional

from .protocol import MERGED_KEY
from .protocol import MLCoreProtocol
from ...types import tensor_dict_type
from ...protocol import TrainerState
from ...constants import PREDICTIONS_KEY
from ...modules.blocks import mapping_dict


@MLCoreProtocol.register("fcnn")
class FCNN(MLCoreProtocol):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_history: int,
        hidden_units: Optional[List[int]] = None,
        *,
        mapping_type: str = "basic",
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
    ):
        super().__init__(in_dim, out_dim, num_history)
        in_dim *= num_history
        if hidden_units is None:
            dim = max(32, min(1024, 2 * in_dim))
            hidden_units = 2 * [dim]
        mapping_base = mapping_dict[mapping_type]
        blocks: List[nn.Module] = []
        for hidden_unit in hidden_units:
            mapping = mapping_base(
                in_dim,
                hidden_unit,
                bias=bias,
                activation=activation,
                batch_norm=batch_norm,
                dropout=dropout,
            )
            blocks.append(mapping)
            in_dim = hidden_unit
        blocks.append(nn.Linear(in_dim, out_dim, bias))
        self.hidden_units = hidden_units
        self.net = nn.Sequential(*blocks)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        net = batch[MERGED_KEY]
        if len(net.shape) > 2:
            net = net.contiguous().view(len(net), -1)
        return {PREDICTIONS_KEY: self.net(net)}


__all__ = ["FCNN"]



from torchvision.models import resnet18
from torchvision.models import resnet50
from torchvision.models import resnet101
from torchvision.models import resnet152

from ..register import register_backbone


register_backbone(
    "resnet18",
    [64, 64, 128, 256, 512],
    {
        "relu": "stage0",
        "layer1": "stage1",
        "layer2": "stage2",
        "layer3": "stage3",
        "layer4": "stage4",
    },
)(resnet18)


register_backbone(
    "resnet50",
    [64, 256, 512, 1024, 2048],
    {
        "relu": "stage0",
        "layer1": "stage1",
        "layer2": "stage2",
        "layer3": "stage3",
        "layer4": "stage4",
    },
)(resnet50)


register_backbone(
    "resnet101",
    [64, 256, 512, 1024, 2048],
    {
        "relu": "stage0",
        "layer1": "stage1",
        "layer2": "stage2",
        "layer3": "stage3",
        "layer4": "stage4",
    },
)(resnet101)


register_backbone(
    "resnet152",
    [64, 256, 512, 1024, 2048],
    {
        "relu": "stage0",
        "layer1": "stage1",
        "layer2": "stage2",
        "layer3": "stage3",
        "layer4": "stage4",
    },
)(resnet152)


__all__ = []  # type: ignore



import torch

import numpy as np
import torch.nn.functional as F

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Optional

from .fcnn import FCNN
from .protocol import MLCoreProtocol
from ..bases import CustomLossBase
from ...types import losses_type
from ...types import tensor_dict_type
from ...protocol import LossProtocol
from ...protocol import TrainerState
from ...constants import LOSS_KEY
from ...constants import INPUT_KEY
from ...constants import LABEL_KEY
from ...constants import PREDICTIONS_KEY
from ..implicit.siren import make_grid
from ..implicit.siren import Siren
from ...misc.toolkit import get_gradient


def all_exists(*tensors: Optional[Tensor]) -> bool:
    return all(tensor is not None for tensor in tensors)


def _expand_element(
    n: int,
    element: Union[float, List[float], np.ndarray, Tensor],
    device: Optional[torch.device] = None,
) -> Tensor:
    if isinstance(element, Tensor):
        return element
    if isinstance(element, np.ndarray):
        element_arr = element.reshape(*element.shape, 1)
    elif isinstance(element, float):
        element_arr = np.repeat(element, n).astype(np.float32)
        element_arr = element_arr.reshape([-1, 1, 1])
    else:
        element_arr = np.array(element, np.float32).reshape([1, -1, 1])
        element_arr = np.repeat(element_arr, n, axis=0)
    element_tensor = torch.from_numpy(element_arr)
    if device is not None:
        element_tensor = element_tensor.to(device)
    return element_tensor


def _make_ddr_grid(num_samples: int, device: torch.device) -> Tensor:
    return make_grid(num_samples + 2, 1, device)[:, 1:-1]


@MLCoreProtocol.register("ddr")
class DDR(CustomLossBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        num_history: int,
        hidden_units: Optional[List[int]] = None,
        *,
        mapping_type: str = "basic",
        bias: bool = True,
        activation: str = "ReLU",
        batch_norm: bool = False,
        dropout: float = 0.0,
        w_sin: float = 1.0,
        w_sin_initial: float = 30.0,
        num_random_samples: int = 16,
        predict_quantiles: bool = True,
        predict_cdf: bool = True,
        y_min_max: Optional[Tuple[float, float]] = None,
    ):
        super().__init__(in_dim, out_dim, num_history)
        self.fcnn = FCNN(
            in_dim,
            out_dim,
            num_history,
            hidden_units,
            mapping_type=mapping_type,
            bias=bias,
            activation=activation,
            batch_norm=batch_norm,
            dropout=dropout,
        )
        hidden_units = self.fcnn.hidden_units
        assert hidden_units is not None
        if not len(set(hidden_units)) == 1:
            raise ValueError("`DDR` requires all hidden units to be identical")

        def _make_siren(_in_dim: Optional[int] = None) -> Siren:
            return Siren(
                None,
                _in_dim or 1,
                out_dim,
                hidden_units[0],  # type: ignore
                num_layers=len(hidden_units),  # type: ignore
                w_sin=w_sin,
                w_sin_initial=w_sin_initial,
                bias=False,
                keep_edge=False,
                use_modulator=False,
            )

        self.predict_quantiles = predict_quantiles
        self.predict_cdf = predict_cdf
        self.q_siren = None if not predict_quantiles else _make_siren()
        self.cdf_siren = None if not predict_cdf else _make_siren(out_dim)
        self.num_random_samples = num_random_samples
        self._y_min_max = y_min_max
        self.register_buffer("y_min_max", torch.tensor([0.0, 0.0]))

    def _init_with_trainer(self, trainer: Any) -> None:
        if self._y_min_max is None:
            y_train = trainer.train_loader.data.y
            self._y_min_max = y_train.min().item(), y_train.max().item()
        self.y_min_max = torch.tensor(self._y_min_max)

    def _get_quantiles(
        self,
        tau: Tensor,
        mods: List[Tensor],
        median: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        q_increment = self.q_siren(mods, init=tau)  # type: ignore
        return q_increment, median[:, None] + q_increment

    def _get_cdf(
        self,
        y_anchor: Tensor,
        median: Tensor,
        y_span: float,
        mods: List[Tensor],
    ) -> Tuple[Tensor, Tensor, Tensor]:
        y_residual = y_anchor - median.unsqueeze(1)
        y_ratio = y_residual / y_span
        logit = self.cdf_siren(mods, init=y_ratio)  # type: ignore
        cdf = torch.sigmoid(logit)
        return y_ratio, logit, cdf

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        # prepare
        net = batch[INPUT_KEY]
        device = net.device
        num_samples = len(net)
        if len(net.shape) > 2:
            net = net.contiguous().view(num_samples, -1)
        get_quantiles = kwargs.get("get_quantiles", True) and self.predict_quantiles
        get_cdf = kwargs.get("get_cdf", True) and self.predict_cdf
        # median forward
        mods = []
        for block in self.fcnn.net:
            net = block(net)
            mods.append(net)
        median = mods.pop()
        results = {PREDICTIONS_KEY: median}
        if not get_quantiles and not get_cdf:
            return results
        y_min, y_max = self.y_min_max.tolist()  # type: ignore
        y_span = y_max - y_min
        # quantile forward
        if get_quantiles:
            tau = kwargs.get("tau", None)
            if tau is not None:
                tau = _expand_element(num_samples, tau, device) * 2.0 - 1.0
            else:
                shape = num_samples, self.num_random_samples, 1
                if self.training:
                    tau = torch.rand(*shape, device=device) * 2.0 - 1.0
                else:
                    tau = _make_ddr_grid(self.num_random_samples, device)
                    tau = tau.repeat(num_samples, 1, 1)
            tau.requires_grad_(True)
            q_increment, quantiles = self._get_quantiles(tau, mods, median)
            results.update(
                {
                    "tau": tau,
                    "q_increment": q_increment,
                    "quantiles": quantiles,
                }
            )
        # cdf forward
        if get_cdf:
            y_anchor = kwargs.get("y_anchor", None)
            if y_anchor is not None:
                y_anchor = _expand_element(num_samples, y_anchor, device)
            else:
                shape = num_samples, self.num_random_samples, 1
                if self.training:
                    y_anchor = torch.rand(*shape, device=device) * y_span + y_min
                else:
                    y_raw_ratio = _make_ddr_grid(self.num_random_samples, device)
                    y_raw_ratio = 0.5 * (y_raw_ratio + 1.0)
                    y_anchor = (y_raw_ratio * y_span + y_min).repeat(num_samples, 1, 1)
            y_anchor.requires_grad_(True)
            y_ratio, logit, cdf = self._get_cdf(y_anchor, median, y_span, mods)
            pdf = get_gradient(cdf, y_anchor, True, True)  # type: ignore
            results.update(
                {
                    "y_anchor": y_anchor,
                    "logit": logit,
                    "cdf": cdf,
                    "pdf": pdf,
                }
            )
        # dual forward
        if get_quantiles and get_cdf:
            dual_y = results["quantiles"].detach()
            results["dual_cdf"] = self._get_cdf(dual_y, median, y_span, mods)[-1]
        return results

    def _get_losses(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: Any,
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> Tuple[tensor_dict_type, tensor_dict_type]:
        state = trainer.state
        forward_results = self(batch_idx, batch, state, **forward_kwargs)
        if not isinstance(trainer.loss, DDRLoss):
            raise ValueError("`DDR` only supports `DDRLoss` as its loss")
        loss_dict = trainer.loss(forward_results, batch, state, **loss_kwargs)
        return forward_results, loss_dict


@LossProtocol.register("ddr")
class DDRLoss(LossProtocol):
    def _init_config(self) -> None:
        self.lb_ddr = self.config.setdefault("lb_ddr", 1.0)
        self.lb_dual = self.config.setdefault("lb_dual", 1.0)
        self.lb_monotonous = self.config.setdefault("lb_monotonous", 1.0)

    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        tau = forward_results.get("tau")
        quantiles = forward_results.get("quantiles")
        q_increment = forward_results.get("q_increment")
        cdf = forward_results.get("cdf")
        pdf = forward_results.get("pdf")
        logit = forward_results.get("logit")
        y_anchor = forward_results.get("y_anchor")
        dual_cdf = forward_results.get("dual_cdf")
        median = forward_results[PREDICTIONS_KEY]
        labels = batch[LABEL_KEY]
        losses = {}
        weighted_losses = []
        # mae
        mae = F.l1_loss(median, labels, reduction="none")
        losses["mae"] = mae
        weighted_losses.append(mae)
        # quantiles
        labels = labels[:, None]
        if all_exists(tau, quantiles, q_increment):
            quantile_error = labels - quantiles
            tau_raw = 0.5 * (tau.detach() + 1.0)  # type: ignore
            neg_errors = tau_raw * quantile_error
            pos_errors = (tau_raw - 1.0) * quantile_error
            q_loss = torch.max(neg_errors, pos_errors).mean((1, 2), keepdim=True)
            g_tau = get_gradient(q_increment, tau, retain_graph=True, create_graph=True)  # type: ignore
            g_tau_loss = F.relu(-g_tau, inplace=True).mean((1, 2), keepdim=True)  # type: ignore
            losses["q"] = q_loss
            losses["g_tau"] = g_tau_loss
            weighted_losses.append(self.lb_ddr * q_loss)
            weighted_losses.append(self.lb_monotonous * g_tau_loss)
        # cdf
        if all_exists(cdf, pdf, logit, y_anchor):
            indicative = (labels <= y_anchor).to(torch.float32)
            cdf_loss = -indicative * logit + F.softplus(logit)
            cdf_loss = cdf_loss.mean((1, 2), keepdim=True)
            pdf_loss = F.relu(-pdf, inplace=True).mean((1, 2), keepdim=True)  # type: ignore
            losses["cdf"] = cdf_loss
            losses["pdf"] = pdf_loss
            weighted_losses.append(self.lb_ddr * cdf_loss)
            weighted_losses.append(self.lb_monotonous * pdf_loss)
        # dual
        if all_exists(dual_cdf):
            tau_raw = 0.5 * (tau.detach() + 1.0)  # type: ignore
            tau_recover_loss = F.l1_loss(tau_raw, dual_cdf)  # type: ignore
            losses["tau_recover"] = tau_recover_loss
            weighted_losses.append(self.lb_dual * tau_recover_loss)
        # aggregate
        losses[LOSS_KEY] = sum(weighted_losses)  # type: ignore
        return losses


__all__ = [
    "DDR",
    "DDRLoss",
]



from typing import Any
from typing import Optional

from ...data import MLLoader
from ...protocol import LossProtocol
from ...protocol import TrainerState
from ...protocol import MetricProtocol
from ...protocol import InferenceOutputs
from ...protocol import InferenceProtocol


class MLInference(InferenceProtocol):
    def get_outputs(  # type: ignore
        self,
        loader: MLLoader,
        *,
        portion: float = 1.0,
        state: Optional[TrainerState] = None,
        metrics: Optional[MetricProtocol] = None,
        loss: Optional[LossProtocol] = None,
        return_outputs: bool = True,
        use_tqdm: bool = False,
        **kwargs: Any,
    ) -> InferenceOutputs:
        kwargs["loader_name"] = loader.name
        return super().get_outputs(
            loader,
            portion=portion,
            state=state,
            metrics=metrics,
            loss=loss,
            return_outputs=return_outputs,
            use_tqdm=use_tqdm,
            **kwargs,
        )


class CVInference(InferenceProtocol):
    pass


class NLPInference(InferenceProtocol):
    pass


__all__ = [
    "MLInference",
    "CVInference",
    "NLPInference",
]



from typing import Any
from typing import Union
from typing import Tuple
from typing import Optional

from .....data import DummyData
from ....zoo.core import DLZoo


class CLIPWithVQGANAligner:
    def __init__(
        self,
        text: str,
        *,
        resolution: Tuple[int, int] = (400, 224),
        condition_path: Optional[str] = None,
        **kwargs: Any,
    ):
        kwargs["model_config"] = {
            "text": text,
            "resolution": resolution,
            "condition_path": condition_path,
        }
        self.m = DLZoo.load_pipeline("multimodal/clip_vqgan_aligner", **kwargs)

    def run(self, *, cuda: Optional[Union[int, str]]) -> None:
        self.m.fit(DummyData(), cuda=cuda)  # type: ignore


__all__ = [
    "CLIPWithVQGANAligner",
]



import torch

from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from cftool.misc import check_requires

from ..decoder import DecoderBase
from ..encoder import EncoderBase
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....protocol import ModelProtocol
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import LATENT_KEY
from ....constants import PREDICTIONS_KEY
from ....modules.blocks import Conv2d
from ....modules.blocks import ChannelPadding


class VQCodebook(nn.Module):
    def __init__(self, num_code: int, code_dimension: int):
        super().__init__()
        self.num_code = num_code
        self.code_dimension = code_dimension
        self.embedding = nn.Embedding(num_code, code_dimension)
        span = 1.0 / num_code
        self.embedding.weight.data.uniform_(-span, span)

    def forward(self, z_e: Tensor) -> Tuple[Tensor, Tensor]:
        z_e = z_e.permute(0, 2, 3, 1).contiguous()

        codebook = self.embedding.weight.detach()
        with torch.no_grad():
            z_e_flattened = z_e.view(-1, self.code_dimension)
            distances = (
                torch.sum(z_e_flattened**2, dim=1, keepdim=True)
                + torch.sum(codebook**2, dim=1)
                - 2.0 * torch.einsum("bd,dn->bn", z_e_flattened, codebook.t())
            )
            indices = torch.argmin(distances, dim=1)

        z_q_flattened = codebook[indices]
        z_q = z_q_flattened.view_as(z_e)
        # VQSTE in one line of code
        z_q = z_e + (z_q - z_e).detach()

        indices = indices.view(*z_q.shape[:-1])
        z_q = z_q.permute(0, 3, 1, 2).contiguous()
        return z_q, indices


@ModelProtocol.register("vq_generator")
class VQGenerator(ModelProtocol):
    def __init__(
        self,
        img_size: int,
        num_code: int,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        *,
        encoder: str,
        decoder: str,
        code_dimension: int = 256,
        latent_channels: int = 256,
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
        latent_padding_channels: Optional[int] = None,
        num_classes: Optional[int] = None,
    ):
        super().__init__()
        self.img_size = img_size
        self.num_code = num_code
        self.in_channels = in_channels
        self.num_classes = num_classes
        self.code_dimension = code_dimension
        self.latent_channels = latent_channels
        if encoder_config is None:
            encoder_config = {}
        if decoder_config is None:
            decoder_config = {}
        # encoder
        encoder_base = EncoderBase.get(encoder)
        if check_requires(encoder_base, "img_size"):
            encoder_config["img_size"] = img_size
        encoder_config["in_channels"] = in_channels
        encoder_config["latent_channels"] = latent_channels
        self.encoder = encoder_base(**encoder_config)
        latent_resolution = self.encoder.latent_resolution(img_size)
        self.latent_resolution = latent_resolution
        # codebook
        self.codebook = VQCodebook(num_code, code_dimension)
        # decoder
        decoder_base = DecoderBase.get(decoder)
        if check_requires(decoder_base, "img_size"):
            decoder_config["img_size"] = img_size
        decoder_config["out_channels"] = out_channels or in_channels
        decoder_config["latent_resolution"] = latent_resolution
        decoder_config["latent_channels"] = latent_channels
        decoder_config["num_classes"] = num_classes
        self.decoder = decoder_base(**decoder_config)
        # latent padding
        if latent_padding_channels is None:
            self.latent_padding = None
        else:
            self.latent_padding = ChannelPadding(
                latent_channels,
                latent_padding_channels,
                latent_resolution,
            )

    def to_codebook(self, latent: Tensor) -> Tensor:
        return latent

    def from_codebook(self, z_q: Tensor) -> Tensor:
        return z_q

    def encode(self, net: Tensor) -> Tensor:
        net = self.encoder.encode({INPUT_KEY: net})
        net = self.to_codebook(net)
        net, _ = self.codebook(net)
        return net

    def decode(
        self,
        z_q: Tensor,
        *,
        labels: Optional[Tensor] = None,
        resize: bool = True,
    ) -> Tensor:
        if labels is None and self.num_classes is not None:
            labels = torch.randint(self.num_classes, [len(z_q)], device=z_q.device)
        if self.latent_padding is not None:
            z_q = self.latent_padding(z_q)
        net = self.from_codebook(z_q)
        batch = {INPUT_KEY: net, LABEL_KEY: labels}
        net = self.decoder.decode(batch, resize=resize)
        return net

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        z_e = self.encoder(batch_idx, batch, state, **kwargs)[LATENT_KEY]
        net = self.to_codebook(z_e)
        z_q, indices = self.codebook(net)
        net = self.decode(z_q, labels=batch.get(LABEL_KEY))
        return {PREDICTIONS_KEY: net, "z_q": z_q, "z_e": z_e, "indices": indices}

    def get_code_indices(self, net: Tensor, **kwargs: Any) -> Tensor:
        z_e = self.encoder.encode({INPUT_KEY: net}, **kwargs)
        _, indices = self.codebook(z_e)
        return indices

    def get_code(self, code_indices: Tensor) -> Tensor:
        code_indices = code_indices.squeeze(1)
        z_q = self.codebook.embedding(code_indices.to(self.device))
        return z_q.permute(0, 3, 1, 2)

    def reconstruct_from(
        self,
        code_indices: Tensor,
        *,
        class_idx: Optional[int] = None,
        labels: Optional[Tensor] = None,
        use_one_hot: bool = False,
        **kwargs: Any,
    ) -> Tensor:
        z_q = self.get_code(code_indices)
        if use_one_hot:
            one_hot = torch.zeros_like(z_q)
            i = int(round(0.5 * z_q.shape[2]))
            j = int(round(0.5 * z_q.shape[3]))
            one_hot[..., i, j] = z_q[..., i, j]
            z_q = one_hot
        if labels is None and class_idx is not None:
            labels = torch.full([len(z_q)], class_idx, device=self.device)
        return self.decode(z_q, labels=labels, **kwargs)

    def sample_codebook(
        self,
        *,
        code_indices: Optional[Tensor] = None,
        num_samples: Optional[int] = None,
        class_idx: Optional[int] = None,
        **kwargs: Any,
    ) -> Tuple[Tensor, Tensor]:
        if code_indices is None:
            if num_samples is None:
                raise ValueError("either `indices` or `num_samples` should be provided")
            code_indices = torch.randint(self.num_code, [num_samples])
        code_indices = code_indices.view(-1, 1, 1, 1)
        resolution = self.latent_resolution
        tiled = code_indices.repeat([1, 1, resolution, resolution])
        if class_idx is not None:
            kwargs["labels"] = torch.full([len(code_indices)], class_idx)
        kwargs.setdefault("use_one_hot", True)
        net = self.reconstruct_from(tiled, **kwargs)
        return net, code_indices


@VQGenerator.register("vqgan_generator")
class VQGANGenerator(VQGenerator):
    def __init__(
        self,
        img_size: int,
        num_code: int = 16384,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        *,
        code_dimension: int = 256,
        latent_channels: int = 256,
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
        latent_padding_channels: Optional[int] = None,
        num_classes: Optional[int] = None,
    ):
        super().__init__(
            img_size,
            num_code,
            in_channels,
            out_channels,
            encoder="vqgan",
            decoder="vqgan",
            code_dimension=code_dimension,
            latent_channels=latent_channels,
            encoder_config=encoder_config,
            decoder_config=decoder_config,
            latent_padding_channels=latent_padding_channels,
            num_classes=num_classes,
        )
        self.q_conv = Conv2d(latent_channels, code_dimension, kernel_size=1)
        self.post_q_conv = Conv2d(code_dimension, latent_channels, kernel_size=1)

    def to_codebook(self, latent: Tensor) -> Tensor:
        return self.q_conv(latent)

    def from_codebook(self, z_q: Tensor) -> Tensor:
        return self.post_q_conv(z_q)


__all__ = [
    "VQGenerator",
    "VQGANGenerator",
]



from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple

from ....data import Transforms
from ....data import InferenceImageFolderData
from ....types import np_dict_type
from ....pipeline import DLPipeline


class FolderInferenceResults(NamedTuple):
    outputs: np_dict_type
    img_paths: List[str]


def predict_folder(
    m: DLPipeline,
    folder: str,
    *,
    batch_size: int,
    num_workers: int = 0,
    transform: Optional[Union[str, List[str], Transforms, Callable]] = None,
    transform_config: Optional[Dict[str, Any]] = None,
    use_tqdm: bool = True,
    **predict_kwargs: Any,
) -> FolderInferenceResults:
    data = InferenceImageFolderData(
        folder,
        batch_size=batch_size,
        num_workers=num_workers,
        transform=transform,
        transform_config=transform_config,
    )
    outputs = m.predict(data, use_tqdm=use_tqdm, **predict_kwargs)
    return FolderInferenceResults(outputs, data.dataset.img_paths)


__all__ = [
    "predict_folder",
]



import os
import json
import torch

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.dist import Parallel
from cftool.misc import shallow_copy_dict
from cftool.misc import get_latest_workplace
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.misc import LoggingMixin

from .task import Task
from ...types import data_type
from ...constants import DATA_CONFIG_FILE
from ...constants import META_CONFIG_NAME
from ...constants import CHECKPOINTS_FOLDER
from ...api.ml.pipeline import CarefreePipeline


def _task(
    task: Task,
    execute: str,
    config_folder: str,
    cuda: Optional[Union[int, str]] = None,
) -> None:
    task.run(execute, config_folder, cuda)


def inject_distributed_tqdm_kwargs(
    i: int,
    num_jobs: int,
    kwargs: Dict[str, Any],
) -> None:
    tqdm_settings = kwargs.setdefault("tqdm_settings", {})
    if tqdm_settings is None:
        kwargs["tqdm_settings"] = tqdm_settings = {}
    tqdm_settings.setdefault("use_tqdm", True)
    tqdm_settings.setdefault("use_step_tqdm", False)
    tqdm_settings.setdefault("in_distributed", True)
    tqdm_settings.setdefault("tqdm_position", i % (num_jobs or 1) + 1)
    tqdm_settings.setdefault("tqdm_desc", f"epoch (task {i})")


class ExperimentResults(NamedTuple):
    workplaces: List[str]
    workplace_keys: List[Tuple[str, str]]
    pipelines: Optional[List[CarefreePipeline]]

    @property
    def pipeline_dict(self) -> Dict[str, CarefreePipeline]:
        if self.pipelines is None:
            raise ValueError("pipelines are not provided")
        return dict(zip(self.workplaces, self.pipelines))

    @property
    def checkpoint_folders(self) -> Dict[str, List[str]]:
        folders: Dict[str, Dict[int, str]] = {}
        for workplace, workplace_key in zip(self.workplaces, self.workplace_keys):
            model, index = workplace_key
            latest_workplace = get_latest_workplace(workplace)
            assert latest_workplace is not None, "internal error occurred"
            checkpoint_folder = os.path.join(latest_workplace, CHECKPOINTS_FOLDER)
            folders.setdefault(model, {})[int(index)] = checkpoint_folder
        return {k: [v[i] for i in range(len(v))] for k, v in folders.items()}


class Experiment(LoggingMixin):
    tasks_folder = "__tasks__"
    default_root_workplace = "__experiment__"

    def __init__(
        self,
        *,
        num_jobs: int = 1,
        use_cuda: bool = True,
        available_cuda_list: Optional[List[int]] = None,
        resource_config: Optional[Dict[str, Any]] = None,
    ):
        use_cuda = use_cuda and torch.cuda.is_available()
        if available_cuda_list is None and not use_cuda:
            available_cuda_list = []
        self.num_jobs = num_jobs
        self.use_cuda = use_cuda
        self.cuda_list = available_cuda_list
        self.resource_config = resource_config or {}
        self.tasks: Dict[Tuple[str, str], Task] = {}
        self.key_indices: Dict[Tuple[str, str], int] = {}
        self.executes: Dict[Tuple[str, str], str] = {}
        self.workplaces: Dict[Tuple[str, str], str] = {}
        self.results: Optional[ExperimentResults] = None

    @staticmethod
    def data_folder(workplace: Optional[str] = None) -> str:
        workplace = workplace or Experiment.default_root_workplace
        return os.path.join(workplace, "__data__")

    @staticmethod
    def dump_data(
        data_folder: str,
        x: data_type,
        y: data_type = None,
        postfix: str = "",
    ) -> None:
        for key, value in zip([f"x{postfix}", f"y{postfix}"], [x, y]):
            if value is None:
                continue
            np.save(os.path.join(data_folder, f"{key}.npy"), value)

    @staticmethod
    def dump_data_bundle(
        x: data_type,
        y: data_type = None,
        x_valid: data_type = None,
        y_valid: data_type = None,
        *,
        workplace: Optional[str] = None,
        data_folder: Optional[str] = None,
    ) -> str:
        if data_folder is None:
            data_folder = Experiment.data_folder(workplace)
        os.makedirs(data_folder, exist_ok=True)
        if not isinstance(x, np.ndarray):
            data_config = {"x": x, "y": y, "x_valid": x_valid, "y_valid": y_valid}
            for k, v in data_config.items():
                if v is not None:
                    assert isinstance(v, str)
                    data_config[k] = os.path.abspath(v)
            with open(os.path.join(data_folder, DATA_CONFIG_FILE), "w") as f:
                json.dump(data_config, f)
        else:
            Experiment.dump_data(data_folder, x, y)
            Experiment.dump_data(data_folder, x_valid, y_valid, "_valid")
        return data_folder

    @staticmethod
    def fetch_data(
        postfix: str = "",
        *,
        workplace: Optional[str] = None,
        data_folder: Optional[str] = None,
    ) -> Tuple[data_type, data_type]:
        if data_folder is None:
            data_folder = Experiment.data_folder(workplace)
        data = []
        for key in [f"x{postfix}", f"y{postfix}"]:
            file = os.path.join(data_folder, f"{key}.npy")
            data.append(None if not os.path.isfile(file) else np.load(file))
        return data[0], data[1]

    @staticmethod
    def workplace(workplace_key: Tuple[str, str], root_workplace: str) -> str:
        workplace = os.path.join(root_workplace, *workplace_key)
        return os.path.abspath(workplace)

    @property
    def num_tasks(self) -> int:
        return len(self.tasks)

    def add_task(
        self,
        x: data_type = None,
        y: data_type = None,
        x_valid: data_type = None,
        y_valid: data_type = None,
        *,
        model: str = "fcnn",
        execute: str = "basic",
        root_workplace: Optional[str] = None,
        workplace_key: Optional[Tuple[str, str]] = None,
        config: Optional[Dict[str, Any]] = None,
        data_folder: Optional[str] = None,
        **task_meta_kwargs: Any,
    ) -> str:
        if workplace_key is None:
            counter = 0
            while True:
                workplace_key = model, str(counter)
                if workplace_key not in self.tasks:
                    break
                counter += 1
        if workplace_key in self.tasks:
            raise ValueError(f"task already exists with '{workplace_key}'")
        root_workplace = root_workplace or self.default_root_workplace
        workplace = self.workplace(workplace_key, root_workplace)
        copied_config = shallow_copy_dict(config or {})
        copied_config["core_name"] = model
        new_idx = len(self.tasks)
        inject_distributed_tqdm_kwargs(new_idx, self.num_jobs, copied_config)
        if data_folder is None and x is not None:
            data_folder = self.dump_data_bundle(x, y, x_valid, y_valid, workplace=workplace)
        if data_folder is not None:
            copied_config["data_folder"] = os.path.abspath(data_folder)
        new_task = Task(
            workplace=workplace,
            config=copied_config,
            **shallow_copy_dict(task_meta_kwargs),
        )
        self.tasks[workplace_key] = new_task
        self.key_indices[workplace_key] = new_idx
        self.executes[workplace_key] = execute
        self.workplaces[workplace_key] = workplace
        return workplace

    def run_tasks(
        self,
        *,
        use_tqdm: bool = True,
        task_loader: Optional[Callable[[str], CarefreePipeline]] = None,
        **parallel_kwargs: Any,
    ) -> ExperimentResults:
        resource_config = shallow_copy_dict(self.resource_config)
        gpu_config = resource_config.setdefault("gpu_config", {})
        gpu_config["available_cuda_list"] = self.cuda_list
        parallel = Parallel(
            self.num_jobs,
            use_tqdm=use_tqdm,
            use_cuda=self.use_cuda,
            resource_config=resource_config,
            **parallel_kwargs,
        )
        sorted_workplace_keys = sorted(self.tasks, key=self.key_indices.get)  # type: ignore
        sorted_tasks = [self.tasks[key] for key in sorted_workplace_keys]
        sorted_executes = [self.executes[key] for key in sorted_workplace_keys]
        sorted_workplaces = [self.workplaces[key] for key in sorted_workplace_keys]
        parallel(_task, sorted_tasks, sorted_executes, sorted_workplaces)
        if task_loader is None:
            pipelines = None
        else:
            pipelines = list(map(task_loader, sorted_workplaces))
        self.results = ExperimentResults(
            sorted_workplaces,
            sorted_workplace_keys,
            pipelines,
        )
        return self.results

    def save(self, export_folder: str, *, compress: bool = True) -> "Experiment":
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            Saving.prepare_folder(self, export_folder)
            # tasks
            tasks_folder = os.path.join(abs_folder, self.tasks_folder)
            for workplace_key, task in self.tasks.items():
                task.save(os.path.join(tasks_folder, *workplace_key))
            # meta
            if self.results is None:
                raise ValueError("results are not generated yet")
            meta_config = {
                "executes": self.executes,
                "num_jobs": self.num_jobs,
                "use_cuda": self.use_cuda,
                "cuda_list": self.cuda_list,
                "resource_config": self.resource_config,
                "results": ExperimentResults(
                    self.results.workplaces,
                    self.results.workplace_keys,
                    None,
                ),
            }
            Saving.save_dict(meta_config, META_CONFIG_NAME, abs_folder)
            if compress:
                Saving.compress(abs_folder, remove_original=True)
        return self

    @classmethod
    def load(
        cls,
        saving_folder: str,
        *,
        compress: bool = True,
        task_loader: Optional[Callable[[str], CarefreePipeline]] = None,
    ) -> "Experiment":
        workplace: str
        workplace_key: Tuple[str, str]

        abs_folder = os.path.abspath(saving_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [saving_folder]):
            with Saving.compress_loader(abs_folder, compress, remove_extracted=False):
                meta_config = Saving.load_dict(META_CONFIG_NAME, abs_folder)
                experiment = cls(
                    num_jobs=meta_config["num_jobs"],
                    use_cuda=meta_config["use_cuda"],
                    available_cuda_list=meta_config["cuda_list"],
                    resource_config=meta_config["resource_config"],
                )
                experiment.executes = meta_config["executes"]
                results = list(meta_config["results"])
                # tasks
                pipelines = []
                experiment.tasks = {}
                tasks_folder = os.path.join(abs_folder, cls.tasks_folder)
                for workplace, workplace_key in zip(*results[:2]):
                    task_folder = os.path.join(tasks_folder, *workplace_key)
                    experiment.tasks[workplace_key] = Task.load(task_folder)
                    if task_loader is not None:
                        pipelines.append(task_loader(workplace))
                results[-1] = pipelines or None
                experiment.results = ExperimentResults(*results)
        return experiment


__all__ = [
    "inject_distributed_tqdm_kwargs",
    "Experiment",
    "ExperimentResults",
]



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import Callable
from typing import Optional
from transformers import AutoModel
from transformers import AutoTokenizer
from cftool.misc import check_requires
from cftool.misc import shallow_copy_dict

from ...bases import ModelProtocol
from ....types import texts_type
from ....types import np_dict_type
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....protocol import InferenceProtocol
from ....data.interface import TensorDictData


@ModelProtocol.register("hugging_face")
class HuggingFaceModel(ModelProtocol):
    tokenizer_base: Type = AutoTokenizer
    model_base: Type = AutoModel
    forward_fn_name: str = "forward"

    def __init__(self, model: str):
        super().__init__()
        self.tokenizer = self.tokenizer_base.from_pretrained(model)
        self.model = self.model_base.from_pretrained(model)

    def model_forward(self, batch: tensor_dict_type) -> Any:
        fn = getattr(self.model, self.forward_fn_name)
        batch = shallow_copy_dict(batch)
        pop_keys = [k for k in batch.keys() if not check_requires(fn, k, False)]
        for k in pop_keys:
            batch.pop(k)
        return fn(**batch)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        return self.model_forward(batch)

    def inference(
        self,
        texts: texts_type,
        use_tqdm: bool = True,
        **kwargs: Any,
    ) -> np_dict_type:
        x = self.tokenizer(texts, padding=True, return_tensors="pt")
        data = TensorDictData(x)
        data.prepare(None)
        loader = data.initialize()[0]
        inference = InferenceProtocol(model=self)
        outputs = inference.get_outputs(loader, use_tqdm=use_tqdm, **kwargs)
        return outputs.forward_results

    def to_onnx(  # type: ignore
        self,
        export_folder: str,
        sample_text: str,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        max_length: Optional[int] = None,
        onnx_file: str = "model.onnx",
        opset: int = 11,
        simplify: bool = True,
        forward_fn: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        output_names: Optional[List[str]] = None,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> "ModelProtocol":
        if max_length is None:
            input_sample = self.tokenizer(sample_text, return_tensors="pt")
        else:
            input_sample = self.tokenizer(
                sample_text,
                padding="max_length",
                truncation=True,
                max_length=max_length,
                return_tensors="pt",
            )
        return super().to_onnx(
            export_folder,
            input_sample,
            dynamic_axes,
            onnx_file=onnx_file,
            opset=opset,
            simplify=simplify,
            forward_fn=forward_fn,
            output_names=output_names,
            num_samples=num_samples,
            verbose=verbose,
            **kwargs,
        )


__all__ = [
    "HuggingFaceModel",
]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cftool.misc import check_requires

from .encoder import EncoderBase
from .encoder import Encoder1DBase
from .decoder import DecoderBase
from .decoder import Decoder1DBase
from ...types import tensor_dict_type
from ...misc.toolkit import auto_num_layers


class EncoderDecoder(nn.Module):
    def __init__(
        self,
        is_1d: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        target_downsample: Optional[int] = None,
        latent_padding_channels: Optional[int] = 16,
        num_classes: Optional[int] = None,
        *,
        latent: int = 128,
        decoder_latent: Optional[int] = None,
        img_size: Optional[int] = None,
        min_size: int = 2,
        num_downsample: Optional[int] = None,
        num_upsample: Optional[int] = None,
        latent_resolution: Optional[int] = None,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        # up / down sample stuffs
        if num_downsample is None:
            if img_size is None:
                raise ValueError(
                    "either `img_size` or `num_downsample` should be provided "
                    "(when `num_downsample` is not provided, it will be inferred "
                    "automatically with `img_size` & `min_size`)"
                )
            args = img_size, min_size, target_downsample
            num_downsample = auto_num_layers(*args, use_stride=encoder == "vanilla")
        num_upsample = num_upsample or num_downsample
        if latent_resolution is None and img_size is None:
            raise ValueError(
                "either `img_size` or `latent_resolution` should be provided "
                "in `VanillaVAEBase` (when `latent_resolution` is not provided, "
                "it will be inferred automatically with `img_size`)"
            )
        # check num_downsample requirement
        self.encoder_base = (Encoder1DBase if is_1d else EncoderBase).get(encoder)  # type: ignore
        self.decoder_base = (Decoder1DBase if is_1d else DecoderBase).get(decoder)  # type: ignore
        requires_num_downsample = check_requires(self.encoder_base, "num_downsample")
        if not requires_num_downsample:
            num_downsample = None
        # properties
        self.is_1d = is_1d
        self.latent = latent
        self.decoder_latent = decoder_latent or latent
        self.latent_key = "latent_dim" if is_1d else "latent_channels"
        self.img_size = img_size
        self.latent_padding_channels = latent_padding_channels
        self.num_classes = num_classes
        self.num_downsample = num_downsample
        self.num_upsample = num_upsample
        # encoder
        if encoder_config is None:
            encoder_config = {}
        if check_requires(self.encoder_base, "img_size"):
            encoder_config["img_size"] = img_size
        if requires_num_downsample:
            encoder_config["num_downsample"] = num_downsample
        encoder_config["in_channels"] = in_channels
        encoder_config[self.latent_key] = latent
        self.encoder = self.encoder_base.make(encoder, config=encoder_config)
        if not is_1d and img_size is not None:
            latent_resolution = self.encoder.latent_resolution(img_size)
        # decoder
        if decoder_config is None:
            decoder_config = {}
        decoder_config[self.latent_key] = self.decoder_latent
        decoder_config["img_size"] = img_size
        decoder_config["num_classes"] = num_classes
        decoder_config["out_channels"] = out_channels or in_channels
        if num_upsample is not None and not decoder.startswith("style"):
            decoder_config["num_upsample"] = num_upsample
        if latent_resolution is not None:
            decoder_config["latent_resolution"] = latent_resolution
        self.decoder = self.decoder_base.make(decoder, config=decoder_config)
        self.latent_resolution = self.decoder.latent_resolution

    def resize(self, net: Tensor) -> Tensor:
        return self.decoder.resize(net)

    def encode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        return self.encoder.encode(batch, **kwargs)

    def decode(self, batch: tensor_dict_type, **kwargs: Any) -> Tensor:
        net = self.decoder.decode(batch, **kwargs)
        return self.resize(net)


__all__ = ["EncoderDecoder"]



import torch

import torch.nn.functional as F

from typing import Any
from typing import Optional

from ..types import losses_type
from ..types import tensor_dict_type
from ..protocol import LossProtocol
from ..protocol import TrainerState
from ..constants import LOSS_KEY
from ..constants import INPUT_KEY
from ..constants import PREDICTIONS_KEY


MU_KEY = "mu"
LOG_VAR_KEY = "log_var"


@LossProtocol.register("vae")
@LossProtocol.register("vae1d")
@LossProtocol.register("vae2d")
@LossProtocol.register("siren_vae")
@LossProtocol.register("style_vae")
class VAELoss(LossProtocol):
    kld_w: torch.Tensor

    def _init_config(self) -> None:
        self.kld_ema = self.config.setdefault("kld_ema", 0.999)
        self.kld_weight = self.config.setdefault("kld_weight", 1.0e-3)
        self.register_buffer("kld_w", torch.tensor([0.0], dtype=torch.float32))

    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        # kld ratio
        if state is not None and not state.is_terminate and self.training:
            eta = self.kld_ema
            ratio = (state.step % state.num_step_per_epoch) / state.num_step_per_epoch
            ratio = (1.0 + ratio) * self.kld_weight
            self.kld_w = eta * self.kld_w + (1.0 - eta) * ratio
        # reconstruction loss
        original = batch[INPUT_KEY]
        reconstruction = forward_results[PREDICTIONS_KEY]
        mse = F.mse_loss(reconstruction, original)
        # kld loss
        mu, log_var = map(forward_results.get, [MU_KEY, LOG_VAR_KEY])
        assert mu is not None and log_var is not None
        var = log_var.exp()
        dim = tuple(i for i in range(1, len(mu.shape)))
        kld_losses = -0.5 * torch.sum(1 + log_var - mu**2 - var, dim=dim)
        kld_loss = torch.mean(kld_losses, dim=0)
        # gather
        if self.training:
            kld_loss = self.kld_w * kld_loss
        loss = mse + kld_loss
        return {"mse": mse, "kld": kld_loss, "kld_w": self.kld_w, LOSS_KEY: loss}


@LossProtocol.register("vq_vae")
class VQVAELoss(LossProtocol):
    def _init_config(self) -> None:
        self.lb_vq = self.config.setdefault("lb_vq", 1.0)
        self.lb_recon = self.config.setdefault("lb_recon", 1.0)
        self.lb_commit = self.config.setdefault("lb_commit", 1.0)

    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        # reconstruction loss
        original = batch[INPUT_KEY]
        reconstruction = forward_results[PREDICTIONS_KEY]
        mse = F.mse_loss(reconstruction, original)
        # vq & commit loss
        z_e = forward_results["z_e"]
        z_q_g = forward_results["z_q_g"]
        vq_loss = F.mse_loss(z_q_g, z_e.detach())
        commit_loss = F.mse_loss(z_e, z_q_g.detach())
        # gather
        loss = self.lb_recon * mse + self.lb_vq * vq_loss + self.lb_commit * commit_loss
        return {"mse": mse, "commit": commit_loss, LOSS_KEY: loss}


__all__ = [
    "MU_KEY",
    "LOG_VAR_KEY",
    "VAELoss",
    "VQVAELoss",
]



import torch

from torch import nn
from torch import Tensor
from cftool.array import l2_normalize
from torchvision.transforms import Resize
from torchvision.transforms import Compose
from torchvision.transforms import ToTensor
from torchvision.transforms import Normalize
from torchvision.transforms import CenterCrop
from torchvision.transforms import InterpolationMode
from cfcv.misc.toolkit import to_rgb

from .protocol import PerceptorProtocol
from ...constants import INPUT_KEY
from ...constants import LATENT_KEY
from ..cv.encoder.transformer import ViTEncoder
from ..nlp.encoder.transformer import TeTEncoder


@PerceptorProtocol.register("clip")
class CLIP(PerceptorProtocol):
    def __init__(
        self,
        img_size: int = 224,
        latent_dim: int = 512,
        *,
        # vision
        in_channels: int = 3,
        vision_latent_expand: float = 1.5,
        vision_patch_size: int = 32,
        vision_num_layers: int = 12,
        vision_num_heads: int = 12,
        # text
        vocab_size: int = 49408,
        context_length: int = 77,
        text_latent_expand: float = 1.0,
        text_num_layers: int = 12,
        text_num_heads: int = 8,
    ):
        super().__init__(img_size, context_length)
        self.vision_latent_dim = int(round(latent_dim * vision_latent_expand))
        feedforward_kwargs = {"activation": "quick_gelu"}
        self.vit = ViTEncoder(
            img_size,
            patch_size=vision_patch_size,
            in_channels=in_channels,
            latent_dim=self.vision_latent_dim,
            to_patches_config={"bias": False},
            num_layers=vision_num_layers,
            norm_kwargs={"eps": 1.0e-5},
            first_norm=nn.LayerNorm(self.vision_latent_dim),
            attention_kwargs={"num_heads": vision_num_heads},
            feedforward_kwargs=feedforward_kwargs,
            norm_after_head=True,
            output_dim=latent_dim,
        )
        self.text_num_layers = text_num_layers
        self.text_latent_dim = int(round(latent_dim * text_latent_expand))
        self.token_embedding = nn.Embedding(vocab_size, self.text_latent_dim)
        self.text_transformer = TeTEncoder(
            self.text_latent_dim,
            use_triu_attn_mask=True,
            num_layers=text_num_layers,
            norm_kwargs={"eps": 1.0e-5},
            attention_kwargs={"num_heads": text_num_heads},
            feedforward_kwargs=feedforward_kwargs,
        )
        projection_shape = self.text_latent_dim, latent_dim
        self.text_projection = nn.Parameter(torch.empty(*projection_shape))

        self.reset_parameters()

    def reset_parameters(self) -> None:
        nn.init.normal_(self.token_embedding.weight, std=0.02)
        text_encoder = self.text_transformer.encoder
        nn.init.normal_(text_encoder.pos_encoding.pos_encoding, std=0.01)
        proj_std = (self.text_latent_dim**-0.5) * ((2 * self.text_num_layers) ** -0.5)
        attn_std = self.text_latent_dim**-0.5
        fc_std = (2 * self.text_latent_dim) ** -0.5
        for block in text_encoder.mixing_blocks:
            attn = block.token_mixing.module.net
            mlp = block.channel_mixing.module.net
            nn.init.normal_(attn.in_w, std=attn_std)
            nn.init.normal_(attn.out_linear.weight, std=proj_std)
            nn.init.normal_(mlp[0].weight, std=fc_std)
            nn.init.normal_(mlp[3].weight, std=proj_std)
        nn.init.normal_(self.text_projection, std=self.text_latent_dim**-0.5)

    def encode_image(self, image: Tensor) -> Tensor:
        net = self.vit(0, {INPUT_KEY: image}, determinate=True)[LATENT_KEY]
        return l2_normalize(net)

    def encode_text(self, text: Tensor) -> Tensor:
        net = self.token_embedding(text)
        net = self.text_transformer(0, {INPUT_KEY: net}, determinate=True)[LATENT_KEY]
        batch_size = net.shape[0]
        if self.training or batch_size > 1:
            net = net[torch.arange(batch_size), text.argmax(dim=-1)]
        else:
            # ONNX compatibility
            net = net[:, torch.nonzero(text)[:, 1][-1]]
        net = net @ self.text_projection
        return l2_normalize(net)

    def get_transform(self) -> Compose:
        return Compose(
            [
                Resize(self.img_size, interpolation=InterpolationMode.BICUBIC),
                CenterCrop(self.img_size),
                to_rgb,
                ToTensor(),
                Normalize(
                    (0.48145466, 0.4578275, 0.40821073),
                    (0.26862954, 0.26130258, 0.27577711),
                ),
            ]
        )


__all__ = ["CLIP"]



import os
import json
import torch

from abc import ABC
from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
from typing import NamedTuple
from cftool.misc import update_dict

from ...types import tensor_dict_type
from ...pipeline import DLPipeline
from ...pipeline import PipelineProtocol
from ...protocol import ModelProtocol
from ...constants import WARNING_PREFIX
from ...constants import DEFAULT_ZOO_TAG
from ...misc.toolkit import inject_debug
from ...misc.toolkit import download_model
from ...misc.toolkit import download_data_info


root = os.path.dirname(__file__)
configs_root = os.path.join(root, "configs")


class ParsedModel(NamedTuple):
    json_path: str
    download_name: str


def _parse_model(model: str) -> ParsedModel:
    tag = DEFAULT_ZOO_TAG
    model_type, model_name = model.split("/")
    download_name = model_name
    if "." in model_name:
        model_name, tag = model_name.split(".")
    json_folder = os.path.join(configs_root, model_type, model_name)
    json_path = os.path.join(json_folder, f"{tag}.json")
    if not os.path.isfile(json_path):
        json_path = os.path.join(json_folder, f"{DEFAULT_ZOO_TAG}.json")
    return ParsedModel(json_path, download_name)


def _parse_config(json_path: str) -> Dict[str, Any]:
    with open(json_path, "r") as f:
        config = json.load(f)
    bases = config.pop("__bases__", [])
    for base in bases[::-1]:
        parsed = _parse_model(base)
        config = update_dict(config, _parse_config(parsed.json_path))
    return config


class ZooBase(ABC):
    def __init__(
        self,
        model: Optional[str] = None,
        *,
        data_info: Optional[Dict[str, Any]] = None,
        json_path: Optional[str] = None,
        no_build: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ):
        self.download_name = None
        # load json
        if json_path is None:
            if model is None:
                raise ValueError("either `model` or `json_path` should be provided")
            parsed = _parse_model(model)
            json_path = parsed.json_path
            self.download_name = parsed.download_name
        self.json_path = json_path
        self.config = _parse_config(json_path)
        if self.download_name is None:
            self.download_name = self.config.pop("tag", None)
        self.err_msg_fmt = f"`{'{}'}` should be provided in '{json_path}'"
        # get pipeline
        self.pipeline_name = self.config.pop("pipeline", None)
        if self.pipeline_name is None:
            raise ValueError(self.err_msg_fmt.format("pipeline"))
        # handle debug
        if debug:
            inject_debug(kwargs)

        # handle requires
        def _inject_requires(d: Dict[str, Any], local_requires: Dict[str, Any]) -> None:
            for k, v in local_requires.items():
                kd = d.setdefault(k, {})
                if isinstance(v, dict):
                    _inject_requires(kd, v)
                    continue
                assert isinstance(v, list), "requirements should be a list"
                for vv in v:
                    if vv in kd:
                        continue
                    required = kwargs.pop(vv, None)
                    if required is None:
                        raise ValueError(f"'{vv}' should be provided in `kwargs`")
                    kd[vv] = required

        requires = self.config.pop("__requires__", {})
        _inject_requires(kwargs, requires)
        # build
        update_dict(kwargs, self.config)
        if no_build:
            self.m = None
        else:
            self.m = DLPipeline.make(self.pipeline_name, self.config)
            if data_info is None:
                if self.download_name is None:
                    data_info = {}
                else:
                    try:
                        with open(download_data_info(self.download_name), "r") as f:
                            data_info = json.load(f)
                    except ValueError:
                        data_info = {}
            try:
                self.m.build(data_info)
            except Exception as err:
                raise ValueError(f"Failed to build '{model}': {err}")

    @classmethod
    def load_pipeline(
        cls,
        model: Optional[str] = None,
        *,
        data_info: Optional[Dict[str, Any]] = None,
        json_path: Optional[str] = None,
        debug: bool = False,
        **kwargs: Any,
    ) -> PipelineProtocol:
        zoo = cls(
            model,
            data_info=data_info,
            json_path=json_path,
            debug=debug,
            **kwargs,
        )
        assert zoo.m is not None
        return zoo.m


class DLZoo(ZooBase):
    m: DLPipeline

    def load_pretrained(self) -> ModelProtocol:
        if self.download_name is None:
            err_msg = self.err_msg_fmt.format("tag")
            raise ValueError(f"{err_msg} when `pretrained` is True")
        m = self.m.model
        m.load_state_dict(torch.load(download_model(self.download_name)))
        return m

    @classmethod
    def load_pipeline(
        cls,
        model: Optional[str] = None,
        *,
        data_info: Optional[Dict[str, Any]] = None,
        json_path: Optional[str] = None,
        pretrained: bool = False,
        debug: bool = False,
        **kwargs: Any,
    ) -> DLPipeline:
        zoo = cls(
            model,
            data_info=data_info,
            json_path=json_path,
            debug=debug,
            **kwargs,
        )
        if pretrained:
            zoo.load_pretrained()
        return zoo.m

    @classmethod
    def load_model(
        cls,
        model: Optional[str] = None,
        *,
        data_info: Optional[Dict[str, Any]] = None,
        json_path: Optional[str] = None,
        pretrained: bool = False,
        **kwargs: Any,
    ) -> ModelProtocol:
        kwargs.setdefault("in_loading", True)
        zoo = cls(model, data_info=data_info, json_path=json_path, **kwargs)
        if pretrained:
            zoo.load_pretrained()
        return zoo.m.model

    @classmethod
    def dump_onnx(
        cls,
        model: str,
        export_folder: str,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        data_info: Optional[Dict[str, Any]] = None,
        json_path: Optional[str] = None,
        onnx_file: str = "model.onnx",
        opset: int = 11,
        simplify: bool = True,
        input_sample: Optional[tensor_dict_type] = None,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> DLPipeline:
        kwargs["in_loading"] = True
        zoo = cls(model, data_info=data_info, json_path=json_path, **kwargs)
        try:
            zoo.load_pretrained()
        except ValueError:
            print(
                f"{WARNING_PREFIX}no pretrained models are available for '{model}', "
                f"so onnx will not be dumped"
            )
            return zoo.m
        zoo.m.to_onnx(
            export_folder,
            dynamic_axes,
            onnx_file=onnx_file,
            opset=opset,
            simplify=simplify,
            input_sample=input_sample,
            num_samples=num_samples,
            verbose=verbose,
        )
        return zoo.m


__all__ = ["DLZoo"]



from .gan import *
from .ssl import *
from .vae import *
from .encoder import *
from .decoder import *
from .stylizer import *
from .generator import *
from .segmentor import *
from .classifier import *



import torch
import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Optional

from .protocol import EncoderBase
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....constants import INPUT_KEY
from ....constants import LATENT_KEY
from ....modules.blocks import Conv2d


def normalize(dim: int) -> nn.GroupNorm:
    return nn.GroupNorm(num_groups=32, num_channels=dim, eps=1.0e-6, affine=True)


class AttnBlock(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.in_channels = dim
        self.norm = normalize(dim)
        self.q = Conv2d(dim, dim, kernel_size=1, stride=1, padding=0)
        self.k = Conv2d(dim, dim, kernel_size=1, stride=1, padding=0)
        self.v = Conv2d(dim, dim, kernel_size=1, stride=1, padding=0)
        self.proj_out = Conv2d(dim, dim, kernel_size=1, stride=1, padding=0)

    def forward(self, net: Tensor) -> Tensor:
        inp = net
        net = self.norm(net)
        q = self.q(net)
        k = self.k(net)
        v = self.v(net)

        b, c, h, w = q.shape
        q = q.reshape(b, c, h * w)
        q = q.permute(0, 2, 1)
        k = k.reshape(b, c, h * w)
        weights = torch.bmm(q, k)
        weights = weights * (int(c) ** (-0.5))
        weights = torch.nn.functional.softmax(weights, dim=2)

        v = v.reshape(b, c, h * w)
        weights = weights.permute(0, 2, 1)
        net = torch.bmm(v, weights)
        net = net.reshape(b, c, h, w)
        net = self.proj_out(net)
        return inp + net


class Downsample(nn.Module):
    def __init__(self, dim: int):
        super().__init__()
        self.net = Conv2d(dim, dim, kernel_size=3, stride=2, padding=0)

    def forward(self, net: Tensor) -> Tensor:
        net = nn.functional.pad(net, (0, 1, 0, 1), mode="constant", value=0.0)
        net = self.net(net)
        return net


class Swish(nn.Module):
    def forward(self, net: Tensor) -> Tensor:
        return net * torch.sigmoid(net)


class ResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: Optional[int] = None,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.in_channels = in_channels
        out_channels = in_channels if out_channels is None else out_channels
        self.out_channels = out_channels
        self.net = nn.Sequential(
            normalize(in_channels),
            Swish(),
            Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
            normalize(out_channels),
            Swish(),
            nn.Dropout(dropout),
            Conv2d(out_channels, out_channels, kernel_size=3, padding=1),
        )
        if in_channels == out_channels:
            self.shortcut = None
        else:
            self.shortcut = Conv2d(in_channels, out_channels, kernel_size=1, padding=0)

    def forward(self, net: Tensor) -> Tensor:
        inp = net
        net = self.net(net)
        if self.shortcut is not None:
            inp = self.shortcut(inp)
        return inp + net


@EncoderBase.register("vqgan")
class VQGANEncoder(EncoderBase):
    def __init__(
        self,
        img_size: int,
        in_channels: int = 3,
        latent_channels: int = 256,
        *,
        channel_multipliers: Optional[List[int]] = None,
        attention_resolutions: Optional[List[int]] = None,
        latent_dim: int = 128,
        res_dropout: float = 0.0,
        num_res_blocks: int = 2,
    ):
        if channel_multipliers is None:
            channel_multipliers = [1, 1, 2, 2, 4]
        if attention_resolutions is None:
            attention_resolutions = [16]
        num_downsample = len(channel_multipliers)
        super().__init__(in_channels, num_downsample, latent_channels)
        self.num_res_blocks = num_res_blocks
        # in conv
        blocks = [Conv2d(in_channels, latent_dim, kernel_size=3, padding=1)]
        # downsample
        current_res = img_size
        in_channel_multipliers = [1] + channel_multipliers
        in_nc = 0
        for i in range(num_downsample):
            i_blocks = []
            in_nc = latent_dim * in_channel_multipliers[i]
            out_nc = latent_dim * channel_multipliers[i]
            for j in range(num_res_blocks):
                i_blocks.append(
                    ResidualBlock(
                        in_nc,
                        out_nc,
                        dropout=res_dropout,
                    )
                )
                in_nc = out_nc
                if current_res in attention_resolutions:
                    i_blocks.append(AttnBlock(in_nc))
            if i != num_downsample - 1:
                i_blocks.append(Downsample(in_nc))
                current_res //= 2
            blocks.append(nn.Sequential(*i_blocks))
        # mid
        blocks.extend(
            [
                ResidualBlock(in_nc, in_nc, res_dropout),
                AttnBlock(in_nc),
                ResidualBlock(in_nc, in_nc, res_dropout),
            ]
        )
        # finalize
        blocks.extend(
            [
                normalize(in_nc),
                Swish(),
                Conv2d(in_nc, latent_channels, kernel_size=3, padding=1),
            ]
        )
        self.net = nn.Sequential(*blocks)

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        return {LATENT_KEY: self.net(batch[INPUT_KEY])}


__all__ = [
    "normalize",
    "Swish",
    "AttnBlock",
    "ResidualBlock",
    "VQGANEncoder",
]



import os

from cftool.misc import LoggingMixin

INFO_PREFIX = LoggingMixin.info_prefix
ERROR_PREFIX = LoggingMixin.error_prefix
WARNING_PREFIX = LoggingMixin.warning_prefix

LOSS_KEY = "loss"
INPUT_KEY = "input"
LATENT_KEY = "latent"
PREDICTIONS_KEY = "predictions"
LABEL_KEY = "labels"
ORIGINAL_LABEL_KEY = "original_labels"
BATCH_INDICES_KEY = "batch_indices"

PT_PREFIX = "model_"
SCORES_FILE = "scores.json"
CHECKPOINTS_FOLDER = "checkpoints"

META_CONFIG_NAME = "__meta__"
DATA_CONFIG_FILE = "__data__.json"
ML_PIPELINE_SAVE_NAME = "ml_pipeline"

DEFAULT_ZOO_TAG = "default"
CACHE_DIR = os.path.join(os.path.expanduser("~"), ".cache", "carefree-learn")
DATA_CACHE_DIR = os.path.join(CACHE_DIR, "data")



import os
import json
import torch

import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from functools import partial
from collections import OrderedDict
from cfdata.tabular import ColumnTypes
from cftool.misc import get_arguments
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.array import softmax
from cftool.array import squeeze
from cftool.array import is_float
from cfml.misc.toolkit import ModelPattern

from ...data import MLData
from ...data import MLLoader
from ...data import MLDataset
from ...data import DLDataModule
from ...types import configs_type
from ...types import np_dict_type
from ...types import states_callback_type
from ...trainer import get_sorted_checkpoints
from ...pipeline import DLPipeline
from ...protocol import InferenceOutputs
from ...constants import PT_PREFIX
from ...constants import SCORES_FILE
from ...constants import PREDICTIONS_KEY
from ...misc.internal_ import MLInference
from ...models.ml.encoders import Encoder
from ...models.ml.protocol import MLModel


@DLPipeline.register("ml.simple")
class SimplePipeline(DLPipeline):
    model: MLModel
    inference: MLInference
    inference_base = MLInference

    encoder: Optional[Encoder]
    numerical_columns_mapping: Dict[int, int]
    categorical_columns_mapping: Dict[int, int]
    use_one_hot: bool
    use_embedding: bool

    def __init__(
        self,
        core_name: str = "fcnn",
        core_config: Optional[Dict[str, Any]] = None,
        *,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        loss_name: str = "auto",
        loss_config: Optional[Dict[str, Any]] = None,
        # encoder
        only_categorical: bool = False,
        encoder_config: Optional[Dict[str, Any]] = None,
        encoding_methods: Optional[Dict[str, List[str]]] = None,
        encoding_configs: Optional[Dict[str, Dict[str, Any]]] = None,
        default_encoding_methods: Optional[List[str]] = None,
        default_encoding_configs: Optional[Dict[str, Any]] = None,
        # trainer
        state_config: Optional[Dict[str, Any]] = None,
        num_epoch: int = 40,
        max_epoch: int = 1000,
        fixed_epoch: Optional[int] = None,
        fixed_steps: Optional[int] = None,
        log_steps: Optional[int] = None,
        valid_portion: float = 1.0,
        amp: bool = False,
        clip_norm: float = 0.0,
        cudnn_benchmark: bool = False,
        metric_names: Optional[Union[str, List[str]]] = None,
        metric_configs: configs_type = None,
        metric_weights: Optional[Dict[str, float]] = None,
        use_losses_as_metrics: Optional[bool] = None,
        loss_metrics_weights: Optional[Dict[str, float]] = None,
        recompute_train_losses_in_eval: bool = True,
        monitor_names: Optional[Union[str, List[str]]] = None,
        monitor_configs: Optional[Dict[str, Any]] = None,
        callback_names: Optional[Union[str, List[str]]] = None,
        callback_configs: Optional[Dict[str, Any]] = None,
        lr: Optional[float] = None,
        optimizer_name: Optional[str] = None,
        scheduler_name: Optional[str] = None,
        optimizer_config: Optional[Dict[str, Any]] = None,
        scheduler_config: Optional[Dict[str, Any]] = None,
        optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
        use_zero: bool = False,
        workplace: str = "_logs",
        finetune_config: Optional[Dict[str, Any]] = None,
        tqdm_settings: Optional[Dict[str, Any]] = None,
        # misc
        in_loading: bool = False,
        pre_process_batch: bool = True,
        num_repeat: Optional[int] = None,
    ):
        self.config = get_arguments()
        super().__init__(
            loss_name=loss_name,
            loss_config=loss_config,
            state_config=state_config,
            num_epoch=num_epoch,
            max_epoch=max_epoch,
            fixed_epoch=fixed_epoch,
            fixed_steps=fixed_steps,
            log_steps=log_steps,
            valid_portion=valid_portion,
            amp=amp,
            clip_norm=clip_norm,
            cudnn_benchmark=cudnn_benchmark,
            metric_names=metric_names,
            metric_configs=metric_configs,
            metric_weights=metric_weights,
            use_losses_as_metrics=use_losses_as_metrics,
            loss_metrics_weights=loss_metrics_weights,
            recompute_train_losses_in_eval=recompute_train_losses_in_eval,
            monitor_names=monitor_names,
            monitor_configs=monitor_configs,
            callback_names=callback_names,
            callback_configs=callback_configs,
            lr=lr,
            optimizer_name=optimizer_name,
            scheduler_name=scheduler_name,
            optimizer_config=optimizer_config,
            scheduler_config=scheduler_config,
            optimizer_settings=optimizer_settings,
            use_zero=use_zero,
            workplace=workplace,
            finetune_config=finetune_config,
            tqdm_settings=tqdm_settings,
            in_loading=in_loading,
        )
        self.core_name = core_name
        self.core_config = core_config or {}
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.only_categorical = only_categorical
        self.encoder = None
        self.encoder_config = encoder_config or {}
        self.encoding_methods = encoding_methods or {}
        self.encoding_configs = encoding_configs or {}
        if default_encoding_methods is None:
            default_encoding_methods = ["embedding"]
        self.default_encoding_methods = default_encoding_methods
        self.default_encoding_configs = default_encoding_configs or {}
        self._pre_process_batch = pre_process_batch
        self._num_repeat = num_repeat

    def _setup_defaults(self, data_info: Dict[str, Any]) -> None:
        self.is_classification = data_info["is_classification"]
        is_reg = not self.is_classification
        if self.input_dim is None:
            self.input_dim = data_info["input_dim"]
        if self.output_dim is None:
            self.output_dim = 1 if is_reg else data_info["num_classes"]
            if self.output_dim is None:
                raise ValueError(
                    "either `MLData` with `cf_data`, or `output_dim`, "
                    "should be provided"
                )
        self.use_auto_loss = self.loss_name == "auto"
        if self.use_auto_loss:
            self.loss_name = "mae" if is_reg else "focal"
        if self.encoder is None:
            self.numerical_columns_mapping = {i: i for i in range(self.input_dim)}
            self.categorical_columns_mapping = {}
            self.use_one_hot = False
            self.use_embedding = False

    def _prepare_modules(self, data_info: Dict[str, Any]) -> None:
        self._setup_defaults(data_info)
        self._prepare_workplace()
        self._prepare_loss()
        assert isinstance(self.input_dim, int)
        assert isinstance(self.output_dim, int)
        self.model = MLModel(
            self.input_dim,
            self.output_dim,
            data_info["num_history"],
            encoder=self.encoder,
            numerical_columns_mapping=self.numerical_columns_mapping,
            categorical_columns_mapping=self.categorical_columns_mapping,
            use_one_hot=self.use_one_hot,
            use_embedding=self.use_embedding,
            only_categorical=self.only_categorical,
            core_name=self.core_name,
            core_config=self.core_config,
            pre_process_batch=self._pre_process_batch,
            num_repeat=self._num_repeat,
        )
        self.inference = MLInference(model=self.model)

    def _prepare_trainer_defaults(self, data_info: Dict[str, Any]) -> None:
        super()._prepare_trainer_defaults(data_info)
        if self.trainer_config["metric_names"] is None and self.use_auto_loss:
            if data_info["is_classification"]:
                self.trainer_config["metric_names"] = ["acc", "auc"]
            else:
                self.trainer_config["metric_names"] = ["mae", "mse"]
        callback_names = self.trainer_config["callback_names"]
        if "_inject_loader_name" not in callback_names:
            callback_names.append("_inject_loader_name")

    def _make_new_loader(  # type: ignore
        self,
        data: MLData,
        batch_size: int,
        **kwargs: Any,
    ) -> MLLoader:
        x = data.x_train
        return MLLoader(MLDataset(x, None), shuffle=False, batch_size=batch_size)

    def to_pattern(
        self,
        *,
        pre_process: Optional[Callable] = None,
        **predict_kwargs: Any,
    ) -> ModelPattern:
        def _predict(x: np.ndarray) -> np.ndarray:
            if pre_process is not None:
                x = pre_process(x)
            predictions = self.predict(x, **predict_kwargs)[PREDICTIONS_KEY]
            if self.is_classification:
                return np.argmax(predictions, axis=1)[..., None]
            return predictions

        def _predict_prob(x: np.ndarray) -> np.ndarray:
            if not self.is_classification:
                msg = "`predict_prob` should not be called in regression tasks"
                raise ValueError(msg)
            if pre_process is not None:
                x = pre_process(x)
            logits = self.predict(x, **predict_kwargs)[PREDICTIONS_KEY]
            return softmax(logits)

        return ModelPattern(
            init_method=lambda: self,
            predict_method=_predict,
            predict_prob_method=_predict_prob,
        )

    @classmethod
    def fuse_multiple(
        cls,
        export_folders: List[str],
        *,
        cuda: Optional[str] = None,
        compress: bool = True,
        states_callback: states_callback_type = None,
        pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        post_callback: Optional[Callable[[DLPipeline, Dict[str, Any]], None]] = None,
    ) -> "SimplePipeline":
        export_folder = export_folders[0]
        base_folder = os.path.dirname(os.path.abspath(export_folder))
        with lock_manager(base_folder, [export_folder]):
            with Saving.compress_loader(export_folder, compress):
                m = cls._load_infrastructure(
                    export_folder,
                    cuda,
                    pre_callback,
                    post_callback,
                )
                assert isinstance(m, SimplePipeline)
                data_info = DLDataModule.load(export_folder)
        m._num_repeat = m.config["num_repeat"] = len(export_folders)
        m._prepare_modules(data_info)
        m.model.to(m.device)
        merged_states: OrderedDict[str, torch.Tensor] = OrderedDict()
        for i, export_folder in enumerate(export_folders):
            base_folder = os.path.dirname(os.path.abspath(export_folder))
            with lock_manager(base_folder, [export_folder]):
                with Saving.compress_loader(export_folder, compress):
                    checkpoints = get_sorted_checkpoints(export_folder)
                    checkpoint_path = os.path.join(export_folder, checkpoints[0])
                    states = torch.load(checkpoint_path, map_location=m.device)
                    current_keys = list(states.keys())
                    for k, v in list(states.items()):
                        split = k.split(".")
                        split.insert(1, str(i))
                        states[".".join(split)] = v
                    for k in current_keys:
                        states.pop(k)
                    if states_callback is not None:
                        states = states_callback(m, states)
                    merged_states.update(states)
        m.model.load_state_dict(merged_states)
        return m

    def re_save(self, export_folder: str, *, compress: bool = True) -> "SimplePipeline":
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            self._save_misc(export_folder)
            file = f"{PT_PREFIX}-1.pt"
            torch.save(self.model.state_dict(), os.path.join(export_folder, file))
            with open(os.path.join(export_folder, SCORES_FILE), "w") as f:
                json.dump({file: 0.0}, f)
            if compress:
                Saving.compress(abs_folder, remove_original=True)
        return self


@DLPipeline.register("ml.carefree")
class CarefreePipeline(SimplePipeline):
    def __init__(
        self,
        core_name: str = "fcnn",
        core_config: Optional[Dict[str, Any]] = None,
        *,
        input_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        loss_name: str = "auto",
        loss_config: Optional[Dict[str, Any]] = None,
        # encoder
        only_categorical: bool = False,
        encoder_config: Optional[Dict[str, Any]] = None,
        encoding_methods: Optional[Dict[str, List[str]]] = None,
        encoding_configs: Optional[Dict[str, Dict[str, Any]]] = None,
        default_encoding_methods: Optional[List[str]] = None,
        default_encoding_configs: Optional[Dict[str, Any]] = None,
        # trainer
        state_config: Optional[Dict[str, Any]] = None,
        num_epoch: int = 40,
        max_epoch: int = 1000,
        fixed_epoch: Optional[int] = None,
        fixed_steps: Optional[int] = None,
        log_steps: Optional[int] = None,
        valid_portion: float = 1.0,
        amp: bool = False,
        clip_norm: float = 0.0,
        cudnn_benchmark: bool = False,
        metric_names: Optional[Union[str, List[str]]] = None,
        metric_configs: configs_type = None,
        metric_weights: Optional[Dict[str, float]] = None,
        use_losses_as_metrics: Optional[bool] = None,
        loss_metrics_weights: Optional[Dict[str, float]] = None,
        recompute_train_losses_in_eval: bool = True,
        monitor_names: Optional[Union[str, List[str]]] = None,
        monitor_configs: Optional[Dict[str, Any]] = None,
        callback_names: Optional[Union[str, List[str]]] = None,
        callback_configs: Optional[Dict[str, Any]] = None,
        lr: Optional[float] = None,
        optimizer_name: Optional[str] = None,
        scheduler_name: Optional[str] = None,
        optimizer_config: Optional[Dict[str, Any]] = None,
        scheduler_config: Optional[Dict[str, Any]] = None,
        optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
        use_zero: bool = False,
        workplace: str = "_logs",
        finetune_config: Optional[Dict[str, Any]] = None,
        tqdm_settings: Optional[Dict[str, Any]] = None,
        # misc
        in_loading: bool = False,
        pre_process_batch: bool = True,
        num_repeat: Optional[int] = None,
    ):
        config = get_arguments()
        super().__init__(
            core_name,
            core_config,
            input_dim=input_dim,
            output_dim=output_dim,
            loss_name=loss_name,
            loss_config=loss_config,
            only_categorical=only_categorical,
            encoder_config=encoder_config,
            encoding_methods=encoding_methods,
            encoding_configs=encoding_configs,
            default_encoding_methods=default_encoding_methods,
            default_encoding_configs=default_encoding_configs,
            state_config=state_config,
            num_epoch=num_epoch,
            max_epoch=max_epoch,
            fixed_epoch=fixed_epoch,
            fixed_steps=fixed_steps,
            log_steps=log_steps,
            valid_portion=valid_portion,
            amp=amp,
            clip_norm=clip_norm,
            cudnn_benchmark=cudnn_benchmark,
            metric_names=metric_names,
            metric_configs=metric_configs,
            metric_weights=metric_weights,
            use_losses_as_metrics=use_losses_as_metrics,
            loss_metrics_weights=loss_metrics_weights,
            recompute_train_losses_in_eval=recompute_train_losses_in_eval,
            monitor_names=monitor_names,
            monitor_configs=monitor_configs,
            callback_names=callback_names,
            callback_configs=callback_configs,
            lr=lr,
            optimizer_name=optimizer_name,
            scheduler_name=scheduler_name,
            optimizer_config=optimizer_config,
            scheduler_config=scheduler_config,
            optimizer_settings=optimizer_settings,
            use_zero=use_zero,
            workplace=workplace,
            finetune_config=finetune_config,
            tqdm_settings=tqdm_settings,
            in_loading=in_loading,
            pre_process_batch=pre_process_batch,
            num_repeat=num_repeat,
        )
        self.config = config

    def _prepare_modules(self, data_info: Dict[str, Any]) -> None:
        self.cf_data = data_info["cf_data"]
        # encoder
        excluded = 0
        numerical_columns_mapping = {}
        categorical_columns_mapping = {}
        categorical_dims = []
        encoding_methods = []
        encoding_configs = []
        true_categorical_columns = []
        use_one_hot = False
        use_embedding = False
        if self.cf_data.is_simplify:
            for idx in range(self.cf_data.processed.x.shape[1]):
                numerical_columns_mapping[idx] = idx
        else:
            ts_indices = self.cf_data.ts_indices
            recognizers = self.cf_data.recognizers
            sorted_indices = [idx for idx in sorted(recognizers) if idx != -1]
            for idx in sorted_indices:
                recognizer = recognizers[idx]
                assert recognizer is not None
                if not recognizer.info.is_valid or idx in ts_indices:
                    excluded += 1
                elif recognizer.info.column_type is ColumnTypes.NUMERICAL:
                    numerical_columns_mapping[idx] = idx - excluded
                else:
                    str_idx = str(idx)
                    categorical_dims.append(recognizer.num_unique_values)
                    idx_encoding_methods = self.encoding_methods.setdefault(
                        str_idx,
                        self.default_encoding_methods,
                    )
                    if isinstance(idx_encoding_methods, str):
                        idx_encoding_methods = [idx_encoding_methods]
                    use_one_hot = use_one_hot or "one_hot" in idx_encoding_methods
                    use_embedding = use_embedding or "embedding" in idx_encoding_methods
                    encoding_methods.append(idx_encoding_methods)
                    encoding_configs.append(
                        self.encoding_configs.setdefault(
                            str_idx,
                            self.default_encoding_configs,
                        )
                    )
                    true_idx = idx - excluded
                    true_categorical_columns.append(true_idx)
                    categorical_columns_mapping[idx] = true_idx
        if not true_categorical_columns:
            encoder = None
        else:
            if self.in_loading:
                loaders = []
            else:
                train_loader, valid_loader = self.data.initialize()
                assert isinstance(train_loader, MLLoader)
                train_loader_copy = train_loader.copy()
                train_loader_copy.disable_shuffle()
                loaders = [train_loader_copy]
                if valid_loader is not None:
                    assert isinstance(valid_loader, MLLoader)
                    loaders.append(valid_loader)
            encoder = Encoder(
                self.encoder_config,
                categorical_dims,
                encoding_methods,  # type: ignore
                encoding_configs,
                true_categorical_columns,
                loaders,
            )
        self.encoder = encoder
        self.use_one_hot = use_one_hot
        self.use_embedding = use_embedding
        self.numerical_columns_mapping = numerical_columns_mapping
        self.categorical_columns_mapping = categorical_columns_mapping
        super()._prepare_modules(data_info)

    def _predict_from_outputs(self, outputs: InferenceOutputs) -> np_dict_type:
        assert self.cf_data is not None
        results = outputs.forward_results
        if self.cf_data.is_clf:
            return results
        fn = partial(self.cf_data.recover_labels, inplace=True)
        recovered = {}
        for k, v in results.items():
            if is_float(v):
                if v.shape[1] == 1:
                    v = fn(v)
                else:
                    v = squeeze(np.apply_along_axis(fn, axis=0, arr=v))
            recovered[k] = v
        return recovered

    def _make_new_loader(  # type: ignore
        self,
        data: MLData,
        batch_size: int,
        **kwargs: Any,
    ) -> MLLoader:
        assert self.cf_data is not None
        return MLLoader(
            MLDataset(*self.cf_data.transform(data.x_train, None, **kwargs).xy),
            shuffle=False,
            batch_size=batch_size,
        )

    @classmethod
    def _load_infrastructure(
        cls,
        export_folder: str,
        cuda: Optional[str],
        pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        post_callback: Optional[Callable[[DLPipeline, Dict[str, Any]], None]] = None,
    ) -> "CarefreePipeline":
        m = super()._load_infrastructure(
            export_folder,
            cuda,
            pre_callback,
            post_callback,
        )
        assert isinstance(m, CarefreePipeline)
        m.cf_data = DLDataModule.load(export_folder)["cf_data"]
        return m

    @classmethod
    def _load_states_callback(cls, m: Any, states: Dict[str, Any]) -> Dict[str, Any]:
        if m.encoder is not None:
            encoder_cache_keys = []
            for key in states:
                if key.startswith("encoder") and key.endswith("cache"):
                    encoder_cache_keys.append(key)
            for key in encoder_cache_keys:
                states.pop(key)
        return states


__all__ = [
    "SimplePipeline",
    "CarefreePipeline",
]



import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from abc import abstractmethod
from abc import ABCMeta
from PIL import Image
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from torchvision import transforms
from cftool.misc import check_requires
from cfcv.misc.toolkit import to_rgb

from .noises import noises
from ..protocol import PerceptorProtocol
from ...bases import CustomLossBase
from ....types import tensor_dict_type
from ....protocol import TrainerState
from ....protocol import ModelProtocol
from ....constants import LOSS_KEY
from ....constants import INPUT_KEY
from ....constants import PREDICTIONS_KEY
from ...cv.encoder import EncoderBase
from ...nlp.tokenizers import TokenizerProtocol
from ....misc.toolkit import interpolate
from ....misc.toolkit import download_model
from ....misc.toolkit import DropNoGradStatesMixin


class CutOuts(nn.Module):
    def __init__(
        self,
        cut_size: int,
        num_cuts: int,
        aspect_ratio: float,
        noise_factor: float = 0.1,
    ):
        super().__init__()
        self.cut_size = cut_size
        self.num_cuts = num_cuts
        self.aspect_ratio = aspect_ratio
        self.num_cuts_zoom = int(2 * num_cuts / 3)
        self.noise_factor = noise_factor

        self.aug_zoom = transforms.Compose(
            [
                transforms.RandomPerspective(distortion_scale=0.4, p=0.7),
                transforms.RandomResizedCrop(
                    size=cut_size,
                    scale=(0.1, 0.75),
                    ratio=(0.85, 1.2),
                ),
                transforms.ColorJitter(0.1, 0.1, 0.1),
            ]
        )

        blocks = []
        if aspect_ratio == 1:
            n_s = 0.95
            n_t = 0.5 * (1.0 - n_s)
            blocks.append(
                transforms.RandomAffine(
                    degrees=0,
                    translate=(n_t, n_t),
                    scale=(n_s, n_s),
                )
            )
        elif aspect_ratio > 1:
            n_s = 1.0 / aspect_ratio
            n_t = 0.5 * (1.0 - n_s)
            blocks.append(
                transforms.RandomAffine(
                    degrees=0,
                    translate=(0, n_t),
                    scale=(0.9 * n_s, n_s),
                )
            )
        else:
            n_s = aspect_ratio
            n_t = 0.5 * (1.0 - n_s)
            blocks.append(
                transforms.RandomAffine(
                    degrees=0,
                    translate=(n_t, 0),
                    scale=(0.9 * n_s, n_s),
                )
            )
        blocks.extend(
            [
                transforms.RandomPerspective(distortion_scale=0.2, p=0.7),
                transforms.CenterCrop(size=cut_size),
                transforms.ColorJitter(0.1, 0.1, 0.1),
            ]
        )
        self.aug_wide = transforms.Compose(blocks)

        self.avg_pool = nn.AdaptiveAvgPool2d((self.cut_size, self.cut_size))
        self.max_pool = nn.AdaptiveMaxPool2d((self.cut_size, self.cut_size))

    def forward(self, net: Tensor) -> Tensor:
        cutouts = []
        for _ in range(self.num_cuts):
            cutout = 0.5 * (self.avg_pool(net) + self.max_pool(net))
            if self.aspect_ratio != 1:
                if self.aspect_ratio > 1:
                    cutout = interpolate(cutout, factor=(1.0, self.aspect_ratio))
                else:
                    cutout = interpolate(cutout, factor=(1.0 / self.aspect_ratio, 1.0))
            cutouts.append(cutout)
        net1 = self.aug_zoom(torch.cat(cutouts[: self.num_cuts_zoom], dim=0))
        net2 = self.aug_wide(torch.cat(cutouts[self.num_cuts_zoom :], dim=0))
        net = torch.cat([net1, net2])
        if self.noise_factor > 0.0:
            factors = net.new_empty([self.num_cuts, 1, 1, 1])
            factors.uniform_(0, self.noise_factor)
            net = net + factors * torch.randn_like(net)
        return net


class Aligner(CustomLossBase, metaclass=ABCMeta):
    perceptor: PerceptorProtocol

    def __init__(
        self,
        perceptor: str,
        generator: str,
        resolution: Tuple[int, int] = (400, 224),
        *,
        num_cuts: int = 36,
        perceptor_config: Optional[Dict[str, Any]] = None,
        generator_config: Optional[Dict[str, Any]] = None,
        perceptor_pretrained_name: Optional[str] = None,
        generator_pretrained_name: Optional[str] = None,
        perceptor_pretrained_path: Optional[str] = None,
        generator_pretrained_path: Optional[str] = None,
    ):
        super().__init__()
        self.resolution = resolution
        if perceptor_pretrained_path is None:
            if perceptor_pretrained_name is not None:
                perceptor_pretrained_path = download_model(perceptor_pretrained_name)
        if generator_pretrained_path is None:
            if generator_pretrained_name is not None:
                generator_pretrained_path = download_model(generator_pretrained_name)
        self.perceptor = PerceptorProtocol.make(perceptor, perceptor_config or {})
        self.generator = ModelProtocol.make(generator, generator_config or {})
        if not hasattr(self.generator, "encode"):
            raise ValueError(
                "`generator` should implement the `encode` method, because "
                "`Aligner` will apply optimization on the 'encoded' latent map"
            )
        self.perceptor.requires_grad_(False).eval()
        self.generator.requires_grad_(False).eval()
        with torch.no_grad():
            if perceptor_pretrained_path is not None:
                self.perceptor.load_state_dict(torch.load(perceptor_pretrained_path))
            if generator_pretrained_path is not None:
                self.generator.load_state_dict(torch.load(generator_pretrained_path))
        aspect_ratio = resolution[0] / resolution[1]
        self.cutouts = CutOuts(self.perceptor.img_size, num_cuts, aspect_ratio)


class ClampWithGrad(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Any:
        inp, inp_min, inp_max = args
        ctx.min = inp_min
        ctx.max = inp_max
        ctx.save_for_backward(inp)
        return inp.clamp(inp_min, inp_max)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        (inp,) = ctx.saved_tensors
        grad = grad_outputs[0]
        return grad * (grad * (inp - inp.clamp(ctx.min, ctx.max)) >= 0), None, None


class Text2ImageAligner(DropNoGradStatesMixin, Aligner, metaclass=ABCMeta):
    vision: Optional[EncoderBase]

    def __init__(
        self,
        perceptor: str,
        generator: str,
        tokenizer: str,
        resolution: Tuple[int, int] = (400, 224),
        *,
        text: str,
        noise: str = "fractal",
        condition_path: Optional[str] = None,
        vision_encoder: Optional[str] = None,
        vision_encoder_config: Optional[Dict[str, Any]] = None,
        vision_encoder_pretrained_name: Optional[str] = None,
        vision_encoder_pretrained_path: Optional[str] = None,
        vision_weight: float = 0.1,
        vision_monitor_step: int = 5,
        tokenizer_config: Optional[Dict[str, Any]] = None,
        num_cuts: int = 36,
        perceptor_config: Optional[Dict[str, Any]] = None,
        generator_config: Optional[Dict[str, Any]] = None,
        perceptor_pretrained_name: Optional[str] = None,
        generator_pretrained_name: Optional[str] = None,
        perceptor_pretrained_path: Optional[str] = None,
        generator_pretrained_path: Optional[str] = None,
    ):
        super().__init__(
            perceptor,
            generator,
            resolution,
            num_cuts=num_cuts,
            perceptor_config=perceptor_config,
            generator_config=generator_config,
            perceptor_pretrained_name=perceptor_pretrained_name,
            generator_pretrained_name=generator_pretrained_name,
            perceptor_pretrained_path=perceptor_pretrained_path,
            generator_pretrained_path=generator_pretrained_path,
        )
        # initialize latent map
        if condition_path is not None:
            img = to_rgb(Image.open(condition_path))
            img = img.resize(resolution, Image.LANCZOS)
        else:
            noise_arr = noises[noise](*resolution)
            img = to_rgb(Image.fromarray(noise_arr))
        img_arr = np.array(img.resize(resolution, Image.LANCZOS))
        img_tensor = torch.from_numpy(img_arr.transpose([2, 0, 1])[None, ...])
        self.initial_img = img_tensor.to(torch.float32) / 255.0
        # we assume that the `generator` requires images between [-1, 1]
        latent = self.generator.encode(self.initial_img * 2.0 - 1.0)
        self.z = nn.Parameter(latent)
        # vision constraints
        self.vision = None
        self.vision_weight = vision_weight
        self.vision_monitor_step = vision_monitor_step
        if vision_encoder is not None:
            if condition_path is None:
                fmt = "`{}` should be provided when `{}` is used"
                raise ValueError(fmt.format("condition_path", "vision_encoder"))
            vepp = vision_encoder_pretrained_path
            if vepp is None:
                if vepp is not None:
                    vepp = download_model(vision_encoder_pretrained_name)
            if vision_encoder_config is None:
                vision_encoder_config = {}
            encoder_base = EncoderBase.get(vision_encoder)
            if check_requires(encoder_base, "img_size"):
                vision_encoder_config["img_size"] = resolution
            vision_encoder_config["in_channels"] = self.initial_img.shape[1]
            self.vision = EncoderBase.make(vision_encoder, vision_encoder_config)
            self.vision.requires_grad_(False).eval()
            if vepp is not None:
                with torch.no_grad():
                    self.vision.load_state_dict(torch.load(vepp))
        # vision latent code
        if self.vision is None:
            self.img_code = None
        else:
            img_code = self.vision.encode({INPUT_KEY: self.initial_img})
            self.register_buffer("img_code", img_code)
        # text latent code
        tokenizer_ins = TokenizerProtocol.make(tokenizer, tokenizer_config or {})
        text_tensor = torch.from_numpy(tokenizer_ins.tokenize(text))
        text_code = self.perceptor.encode_text(text_tensor)
        self.register_buffer("text_code", text_code)

    @abstractmethod
    def generate_raw(self) -> Tensor:
        """should generate an image between [-1, 1]"""

    @abstractmethod
    def perceptor_normalize(self, image: Tensor) -> Tensor:
        pass

    def _get_losses(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: Any,
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> Tuple[tensor_dict_type, tensor_dict_type]:
        state = trainer.state
        forward_results = self(batch_idx, batch, state, **forward_kwargs)
        net = forward_results[PREDICTIONS_KEY]
        cutouts = self.cutouts(net)
        perceptor_net = self.perceptor_normalize(cutouts)
        img_code = self.perceptor.encode_image(perceptor_net)
        img_code = img_code.unsqueeze(1)
        distances = img_code.sub(self.text_code).norm(dim=2)
        loss = align_loss = (2.0 * ((0.5 * distances).arcsin() ** 2)).mean()
        losses = {"align": align_loss}
        if self.vision is not None and state.step % self.vision_monitor_step == 0:
            img_code = self.vision.encode({INPUT_KEY: net})
            vision_loss = losses["vision"] = F.mse_loss(img_code, self.img_code)
            loss = loss + vision_loss * self.vision_weight
        losses[LOSS_KEY] = loss
        return forward_results, losses

    # api

    def generate(self) -> Tensor:
        raw = 0.5 * (self.generate_raw() + 1.0)
        clamped = ClampWithGrad.apply(raw, 0.0, 1.0)
        return clamped

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        return {PREDICTIONS_KEY: self.generate()}


__all__ = ["Text2ImageAligner"]



import torch

import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import List
from typing import Tuple

from .backbone import resnet34_mp
from ...protocol import ImageTranslatorMixin
from .....misc.toolkit import interpolate
from .....misc.internal_ import register_module
from .....modules.blocks import get_conv_blocks
from .....modules.blocks import Conv2d
from .....modules.blocks import SEBlock
from .....modules.blocks import Interpolate
from .....modules.blocks import MaxUnpool2d
from .....modules.blocks import AdaptiveAvgPool2d


def psp_stage(num_channels: int, size: int) -> nn.Sequential:
    prior = AdaptiveAvgPool2d(output_size=size)
    conv = Conv2d(num_channels, num_channels, kernel_size=1, bias=False)
    return nn.Sequential(prior, conv)


class PSPModule(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int = 1024,
        sizes: Tuple[int, ...] = (1, 2, 3, 6),
    ):
        super().__init__()
        stages = [psp_stage(in_channels, size) for size in sizes]
        self.stages = nn.ModuleList(stages)
        self.bottleneck = Conv2d(
            in_channels * (len(sizes) + 1),
            out_channels,
            kernel_size=1,
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, net: Tensor, determinate: bool) -> Tensor:
        kw = dict(anchor=net, mode="bilinear", determinate=determinate)
        priors = [interpolate(stage(net), **kw) for stage in self.stages]
        bottle = self.bottleneck(torch.cat(priors + [net], 1))
        return self.relu(bottle)


def make_se(in_channels: int, reduction: int = 16) -> SEBlock:
    return SEBlock(in_channels, in_channels // reduction, impl="fc")


def fuse(global_logits: Tensor, local_logits: Tensor) -> Tensor:
    values, index = torch.max(global_logits, 1)
    index = index[:, None, :, :].float()
    ### index <===> [0, 1, 2]
    ### bg_mask <===> [1, 0, 0]
    bg_mask = index.clone()
    bg_mask[bg_mask == 2] = 1
    ### trimap_mask <===> [0, 1, 0]
    trimap_mask = index.clone()
    trimap_mask[trimap_mask == 2] = 0
    ### fg_mask <===> [0, 0, 1]
    fg_mask = index.clone()
    fg_mask[fg_mask == 1] = 0
    fg_mask[fg_mask == 2] = 1
    fusion_sigmoid = torch.sigmoid(local_logits) * trimap_mask + fg_mask
    return fusion_sigmoid


class ConvUpPSP(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, factor: float):
        super().__init__()
        self.blocks = nn.ModuleList(
            [
                Conv2d(in_channels, out_channels, kernel_size=3, padding=1),
                nn.BatchNorm2d(out_channels),
                nn.ReLU(inplace=True),
                Interpolate(factor, "bilinear"),
            ]
        )

    def forward(self, net: Tensor, determinate: bool) -> Tensor:
        for block in self.blocks[:-1]:
            net = block(net)
        return self.blocks[-1](net, determinate=determinate)


class DecoderStage(nn.Module):
    def __init__(
        self,
        in_channels: int,
        num_layers: int,
        *,
        reduce_first: bool = True,
        reduce_last: bool = True,
        upsample: bool = True,
        padding: int = 1,
        **kwargs: Any,
    ):
        super().__init__()
        blocks = []
        for i in range(num_layers):
            if i == 0 and reduce_first:
                out_channels = in_channels // 2
            elif i == num_layers - 1 and reduce_last:
                out_channels = in_channels // 2
            else:
                out_channels = in_channels
            blocks.extend(
                get_conv_blocks(
                    in_channels,
                    out_channels,
                    3,
                    stride=1,
                    norm_type="batch",
                    activation="relu",
                    padding=padding,
                    **kwargs,
                )
            )
            in_channels = out_channels
        if upsample:
            blocks.append(Interpolate(2, "bilinear"))
        self.blocks = nn.ModuleList(blocks)

    def forward(self, net: Tensor, determinate: bool) -> Tensor:
        for block in self.blocks:
            kw = {}
            if isinstance(block, Interpolate):
                kw["determinate"] = determinate
            net = block(net, **kw)
        return net


@register_module("aim", pre_bases=[ImageTranslatorMixin])
class AIMNet(nn.Module):
    def __init__(self):  # type: ignore
        super().__init__()
        # encoder
        resnet = resnet34_mp()
        self.encoder0 = nn.Sequential(resnet.conv1, resnet.bn1, resnet.relu)
        self.mp0 = resnet.maxpool1
        self.encoder1 = resnet.layer1
        self.mp1 = resnet.maxpool2
        self.encoder2 = resnet.layer2
        self.mp2 = resnet.maxpool3
        self.encoder3 = resnet.layer3
        self.mp3 = resnet.maxpool4
        self.encoder4 = resnet.layer4
        self.mp4 = resnet.maxpool5
        # decoder
        # global
        self.psp_module = PSPModule(512, 512, (1, 3, 5))
        self.psp4 = ConvUpPSP(512, 256, 2)
        self.psp3 = ConvUpPSP(512, 128, 4)
        self.psp2 = ConvUpPSP(512, 64, 8)
        self.psp1 = ConvUpPSP(512, 64, 16)
        # stage 4d
        self.decoder4_g = DecoderStage(1024, 3)
        self.decoder4_g_se = make_se(256)
        # stage 3d
        self.decoder3_g = DecoderStage(512, 3)
        self.decoder3_g_se = make_se(128)
        # stage 2d
        self.decoder2_g = DecoderStage(256, 3)
        self.decoder2_g_se = make_se(64)
        # stage 1d
        self.decoder1_g = DecoderStage(128, 3, reduce_last=False)
        self.decoder1_g_se = make_se(64)
        # stage 0d
        self.decoder0_g = DecoderStage(128, 2, reduce_last=False)
        self.decoder0_g_spatial = Conv2d(2, 1, kernel_size=7, padding=3)
        self.decoder0_g_se = make_se(64)
        self.decoder_final_g = Conv2d(64, 3, kernel_size=3, padding=1)
        # local
        self.bridge_block = DecoderStage(
            512,
            3,
            reduce_first=False,
            reduce_last=False,
            upsample=False,
            padding=2,
            dilation=2,
        )
        self.max_unpool = MaxUnpool2d(2, 2)
        # stage 4l
        self.decoder4_l = DecoderStage(1024, 3, upsample=False)
        # stage 3l
        self.decoder3_l = DecoderStage(512, 3, upsample=False)
        # stage 2l
        self.decoder2_l = DecoderStage(256, 3, upsample=False)
        # stage 1l
        self.decoder1_l = DecoderStage(128, 3, reduce_last=False, upsample=False)
        # stage 0l
        self.decoder0_l = DecoderStage(128, 2, reduce_last=False, upsample=False)
        self.decoder_final_l = Conv2d(64, 1, kernel_size=3, padding=1)

    def forward(self, net: Tensor, *, determinate: bool = False) -> List[Tensor]:
        # encoder
        e0 = self.encoder0(net)
        # e0: N, 64, H, W
        e0p, id0 = self.mp0(e0)
        e1p, id1 = self.mp1(e0p)
        # e1p: N, 64, H/4, W/4
        e1 = self.encoder1(e1p)
        # e1: N, 64, H/2, W/2
        e2p, id2 = self.mp2(e1)
        # e2p: N, 64, H/8, W/8
        e2 = self.encoder2(e2p)
        # e2: N, 128, H/4, W/4
        e3p, id3 = self.mp3(e2)
        # e3p: N, 128, H/16, W/16
        e3 = self.encoder3(e3p)
        # e3: N, 256, H/8, W/8
        e4p, id4 = self.mp4(e3)
        # e4p: N, 256, H/32, W/32
        e4 = self.encoder4(e4p)
        # e4p: N, 512, H/16, W/16

        # decoder
        psp = self.psp_module(e4, determinate)
        # psp: N, 512, H/32, W/32
        d4_g = torch.cat((psp, e4), 1)
        d4_g = self.decoder4_g(d4_g, determinate)
        # d4_g: N, 256, H/16, W/16
        d4_g = self.decoder4_g_se(d4_g)
        d3_g = torch.cat((self.psp4(psp, determinate), d4_g), 1)
        d3_g = self.decoder3_g(d3_g, determinate)
        # d3_g: N, 128, H/8, W/8
        d3_g = self.decoder3_g_se(d3_g)
        d2_g = torch.cat((self.psp3(psp, determinate), d3_g), 1)
        d2_g = self.decoder2_g(d2_g, determinate)
        # d2_g: N, 64, H/4, W/4
        d2_g = self.decoder2_g_se(d2_g)
        d1_g = torch.cat((self.psp2(psp, determinate), d2_g), 1)
        d1_g = self.decoder1_g(d1_g, determinate)
        # d1_g: N, 64, H/2, W/2
        d1_g = self.decoder1_g_se(d1_g)
        d0_g = torch.cat((self.psp1(psp, determinate), d1_g), 1)
        d0_g = self.decoder0_g(d0_g, determinate)
        # d0_g: N, 64, H, W

        d0_g_avg = torch.mean(d0_g, dim=1, keepdim=True)
        d0_g_max, _ = torch.max(d0_g, dim=1, keepdim=True)
        d0_g_cat = torch.cat([d0_g_avg, d0_g_max], dim=1)
        d0_g_spatial = self.decoder0_g_spatial(d0_g_cat)
        d0_g_spatial_sigmoid = torch.sigmoid(d0_g_spatial)

        d0_g = self.decoder0_g_se(d0_g)
        global_logits = self.decoder_final_g(d0_g)

        bb = self.bridge_block(e4, determinate)
        # bb: N, 512, H/32, W/32
        d4_l = self.decoder4_l(torch.cat((bb, e4), 1), determinate)
        # d4_l: N, 256, H/32, W/32
        d3_l = self.max_unpool(d4_l, id4)
        # d3_l: N, 256, H/16, W/16
        d3_l = self.decoder3_l(torch.cat((d3_l, e3), 1), determinate)
        # d3_l: N, 128, H/16, W/16
        d2_l = self.max_unpool(d3_l, id3)
        # d2_l: N, 128, H/8, W/8
        d2_l = self.decoder2_l(torch.cat((d2_l, e2), 1), determinate)
        # d2_l: N, 64, H/8, W/8
        d1_l = self.max_unpool(d2_l, id2)
        # d1_l: N, 64, H/4, W/4
        d1_l = self.decoder1_l(torch.cat((d1_l, e1), 1), determinate)
        # d1_l: N, 64, H/4, W/4
        d0_l = self.max_unpool(d1_l, id1)
        # d0_l: N, 64, H/2, W/2
        d0_l = self.max_unpool(d0_l, id0)
        # d0_l: N, 64, H, W
        d0_l = self.decoder0_l(torch.cat((d0_l, e0), 1), determinate)
        # d0_l: N, 64, H, W
        d0_l = d0_l + d0_l * d0_g_spatial_sigmoid
        local_logits = self.decoder_final_l(d0_l)

        fusion_sigmoid = fuse(global_logits, local_logits)
        return [fusion_sigmoid, global_logits, local_logits]


__all__ = ["AIMNet"]



import os
import torch
import cflearn

from tqdm import tqdm
from typing import Any
from typing import Optional

from ...interface import load
from ...interface import pack
from ...interface import pixel_cnn
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import INFO_PREFIX
from ....constants import ORIGINAL_LABEL_KEY
from ....data.interface import TensorData
from ....data.interface import CVDataModule
from ....misc.toolkit import to_device
from ....misc.toolkit import save_images
from ....misc.toolkit import eval_context
from ....misc.internal_ import ImageCallback
from ....models.cv.vae.vector_quantized import VQVAE


def register_callback(vqvae: VQVAE, num_classes: Optional[int]) -> None:
    @ImageCallback.register("pixel_cnn")
    class _(ImageCallback):
        def __init__(self, num_keep: int = 25, num_interpolations: int = 16):
            super().__init__(num_keep)
            self.num_interpolations = num_interpolations

        def log_artifacts(self, trainer: cflearn.Trainer) -> None:
            if not self.is_rank_0:
                return None
            device = trainer.device
            batch = next(iter(trainer.validation_loader))
            batch = to_device(batch, device)
            original_indices = batch[cflearn.INPUT_KEY]
            img_size = original_indices.shape[2]
            batch_size = original_indices.shape[0]
            model = trainer.model
            with eval_context(model):
                sampled_indices = model.sample(batch_size, img_size)
            with eval_context(vqvae):
                original = vqvae.reconstruct_from(original_indices.squeeze(1))
                sampled = vqvae.reconstruct_from(sampled_indices.squeeze(1))
            image_folder = self._prepare_folder(trainer)
            save_images(original, os.path.join(image_folder, "original.png"))
            save_images(sampled, os.path.join(image_folder, "sampled.png"))
            if num_classes is not None:
                conditional_folder = os.path.join(image_folder, "conditional")
                os.makedirs(conditional_folder, exist_ok=True)
                for i in range(num_classes):
                    with eval_context(model):
                        sampled_indices = model.sample(batch_size, img_size, i)
                    with eval_context(vqvae):
                        sampled = vqvae.reconstruct_from(sampled_indices, class_idx=i)
                    sampled_path = os.path.join(conditional_folder, f"sampled_{i}.png")
                    save_images(sampled, sampled_path)
                    with eval_context(model):
                        i1 = model.sample(self.num_interpolations, img_size, i)
                        i2 = model.sample(self.num_interpolations, img_size, i)
                    with eval_context(vqvae):
                        z1, z2 = map(vqvae.get_code, [i1, i2])
                        ratio = torch.linspace(
                            0.0,
                            1.0,
                            self.num_interpolations,
                            device=device,
                        )
                        ratio = ratio.view(-1, 1, 1, 1)
                        z_q = ratio * z1 + (1.0 - ratio) * z2
                        shape = [self.num_interpolations, 1]
                        labels = torch.full(shape, i, device=device)
                        interpolations = vqvae.decode(z_q, labels=labels)
                    save_images(
                        interpolations,
                        os.path.join(conditional_folder, f"interpolation_{i}.png"),
                    )


class VQVAEInference:
    def __init__(
        self,
        workplace: str,
        *,
        num_codes: int,
        vqvae_log_folder: str,
        num_classes: Optional[int] = None,
        inference_model: str = "pixel_cnn",
        cuda: Optional[int] = None,
        **kwargs: Any,
    ):
        self.cuda = cuda
        self.debug = kwargs.get("debug", False)
        self.vqvae = load(pack(vqvae_log_folder), cuda=cuda).model
        self.code_export_folder = os.path.join(workplace, "codes")
        register_callback(self.vqvae, num_classes)
        if inference_model == "pixel_cnn":
            model_config = kwargs.setdefault("model_config", {})
            model_config["need_embedding"] = True
            model_config["num_conditional_classes"] = num_classes
            kwargs.setdefault("workplace", os.path.join(workplace, "_logs"))
            self.m = pixel_cnn(num_codes, **kwargs)
        else:
            msg = f"inference model '{inference_model}' is not implemented"
            raise NotImplementedError(msg)

    def export_code_indices(self, data: CVDataModule, export_folder: str) -> None:
        os.makedirs(export_folder, exist_ok=True)
        finished_path = os.path.join(export_folder, "__finished__")
        if os.path.isfile(finished_path):
            return None
        data.prepare(None)
        code_train, code_valid = data.initialize()
        for name, loader in zip(["train", "valid"], [code_train, code_valid]):
            if loader is None:
                continue
            labels = []
            code_indices = []
            for batch in tqdm(loader, desc=f"{name} codes", total=len(loader)):
                labels.append(batch[LABEL_KEY])
                net = batch[INPUT_KEY].to(self.vqvae.device)
                code_indices.append(self.vqvae.get_code_indices(net).unsqueeze(1))
                if self.debug:
                    break
            all_codes = torch.cat(code_indices, dim=0)
            path = os.path.join(export_folder, f"{name}.pt")
            print(f"{INFO_PREFIX}saving {name} codes ({all_codes.shape})")
            torch.save(all_codes, path)
            all_labels = torch.cat(labels, dim=0)
            label_path = os.path.join(export_folder, f"{name}_labels.pt")
            print(f"{INFO_PREFIX}saving {name} labels ({all_labels.shape})")
            torch.save(all_labels, label_path)
        if not self.debug:
            with open(finished_path, "w"):
                pass

    def fit(self, data: CVDataModule) -> "VQVAEInference":
        export_folder = self.code_export_folder
        self.export_code_indices(data, export_folder)
        x_train = torch.load(os.path.join(export_folder, "train.pt"))
        y_train = torch.load(os.path.join(export_folder, "train_labels.pt"))
        x_valid = torch.load(os.path.join(export_folder, "valid.pt"))
        y_valid = torch.load(os.path.join(export_folder, "valid_labels.pt"))
        tensor_data = TensorData(
            x_train,
            y_train=x_train,
            x_valid=x_valid,
            y_valid=x_valid,
            train_others={ORIGINAL_LABEL_KEY: y_train},
            valid_others={ORIGINAL_LABEL_KEY: y_valid},
        )
        self.m.fit(tensor_data, cuda=self.cuda)
        return self


__all__ = [
    "VQVAEInference",
]
